from .util import (
    AxisSelectionEmbeddedTransform,
    HomogeneousEmbeddedTransform,
)
from .base_transform import Transform, InverseTransform, DependentTransformError
from .composite import CompositeTransform, SimplifiedCompositeTransform
from .linear import (
    NullTransform,
    TTransform,
    STTransform,
    AffineTransform,
    SRT3DTransform,
    TransposeTransform,
    PerspectiveTransform,
    BilinearTransform,
    Homography2DTransform,
)
from .nonlinear import LogTransform, PolarTransform, LensDistortionTransform
from .coordinates import Point, PointArray, Vector, VectorArray
from .image import Image
from .systems import CoordinateSystem, CoordinateSystemGraph

__version__ = "2.0.0"


def transform_types():
    typs = [Transform]
    i = 0
    while i < len(typs):
        typs.extend(typs[i].__subclasses__())
        i += 1
    return typs[1:]


_cached_types = None


def create_transform(type, **kwargs):
    global _cached_types
    if _cached_types is None or type not in _cached_types:
        _cached_types = {tr.__name__: tr for tr in transform_types()}

    if type not in _cached_types:
        raise TypeError(f"Unknown transform type {type!r}")

    return _cached_types[type].from_state(kwargs)



import numpy as np
import numpy.linalg
import scipy.optimize

from . import matrices
from ._types import Dims
from .base_transform import Transform
from .params import ArrayParameter, TupleParameter, FloatParameter, TransformParameter


class NullTransform(Transform):
    """Transform having no effect on coordinates (identity transform)."""

    Linear = True
    Orthogonal = True
    NonScaling = True
    Isometric = True
    Equidimensional = True

    def _map(self, coords):
        """Return the input array unmodified.

        Parameters
        ----------
        coords : array-like
            Coordinates to map.
        """
        return coords

    def _imap(self, coords):
        """Return the input array unmodified.

        Parameters
        ----------
        coords : array-like
            Coordinates to inverse map.
        """
        return coords

    def as_affine(self):
        return AffineTransform(
            matrix=np.eye(self.dims[0]),
            offset=np.zeros(self.dims[0]),
            from_cs=self.systems[0],
            to_cs=self.systems[1],
        )

    @property
    def full_matrix(self):
        return np.eye(self.dims[0] + 1)

    def __mul__(self, tr):
        from coorx import CompositeTransform

        if isinstance(tr, CompositeTransform):
            return tr.__rmul__(self)
        if isinstance(tr.inverse, CompositeTransform):
            return tr.inverse.__mul__(self.inverse).inverse
        self.validate_transform_for_mul(tr)
        return tr.copy(from_cs=tr.systems[0], to_cs=self.systems[1])

    def __rmul__(self, tr):
        from coorx import CompositeTransform

        if isinstance(tr, CompositeTransform):
            return tr.__mul__(self)
        if isinstance(tr.inverse, CompositeTransform):
            return tr.inverse.__rmul__(self.inverse).inverse
        tr.validate_transform_for_mul(self)
        return tr.copy(from_cs=self.systems[0], to_cs=tr.systems[1])


class TransposeTransform(Transform):
    """Transposes columns of input coordinates (as opposed to dimensions)."""

    Linear = True
    Orthogonal = True
    NonScaling = False
    Isometric = True
    Equidimensional = True

    parameter_spec = [
        TupleParameter("axis_order", length='dims0'),
    ]

    def __init__(self, axis_order: None | tuple[int, ...] = None, *args, **kwargs):
        super().__init__(*args, **kwargs, axis_order=axis_order)

    @property
    def axis_order(self):
        return self._state['axis_order']

    @axis_order.setter
    def axis_order(self, order):
        self.set_params(axis_order=order)

    def _map(self, coords):
        """Return the input array with columns swapped."""
        return coords[..., self.axis_order]

    def _imap(self, coords):
        """Return the input array columns inversely swapped."""
        return coords[..., np.argsort(self.axis_order)]

    @property
    def params(self):
        return {'axis_order': self.axis_order}

    def as_affine(self):
        return AffineTransform(
            matrix=np.eye(self.dims[0])[:, self.axis_order],
            offset=np.zeros(self.dims[0]),
            from_cs=self.systems[0],
            to_cs=self.systems[1],
        )

    def __rmul__(self, tr):
        if isinstance(tr, TransposeTransform):
            tr.validate_transform_for_mul(self)
            return TransposeTransform(
                axis_order=tuple(self._map(np.array(tr.axis_order))),
                from_cs=self.systems[0],
                to_cs=tr.systems[1],
            )
        return super().__rmul__(tr)


class TTransform(Transform):
    """Transform performing only translation.

    Input/output dimensionality of this transform may be set by the length of the offset parameter.

    Parameters
    ----------
    offset : array-like
        Translation distances.
    dims : tuple
        (input, output) dimensions for transform.
    """

    Linear = True
    Orthogonal = True
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        ArrayParameter(
            "offset",
            dtype=float,
            shape=('dims0',),
            default=0,
        ),
    ]

    def __init__(self, offset=None, dims: Dims = None, **kwargs):
        super().__init__(dims, offset=offset, **kwargs)

    def _map(self, coords):
        """Return translated coordinates.

        Parameters
        ----------
        coords : array-like
            Coordinates to map. The length of the last array dimenstion
            must be equal to the input dimensionality of the transform.

        Returns
        -------
        coords : ndarray
            Mapped coordinates: coords + translation
        """
        return coords + self._state["offset"][np.newaxis, :]

    def _imap(self, coords):
        """Return inverse-mapped coordinates.

        Parameters
        ----------
        coords : array-like
            Coordinates to inverse map. The length of the last array dimenstion
            must be equal to the input dimensionality of the transform.

        Returns
        -------
        coords : ndarray
            Mapped coordinates: coords - translation
        """
        return coords - self._state["offset"][np.newaxis, :]

    @property
    def offset(self):
        return self._state['offset'].copy()

    @offset.setter
    def offset(self, t):
        self.set_params(offset=t)

    def translate(self, offset):
        """Change the translation of this transform by the amount given.

        Parameters
        ----------
        offset : array-like
            The values to be added to the current translation of the transform.
        """
        offset = np.asarray(offset)
        self.offset = self.offset + offset

    def as_affine(self):
        m = AffineTransform(dims=self.dims, from_cs=self.systems[0], to_cs=self.systems[1])
        m.translate(self.offset)
        return m

    def as_st(self):
        return STTransform(
            offset=self.offset,
            scale=(1,) * self.dims[0],
            from_cs=self.systems[0],
            to_cs=self.systems[1],
        )

    def __mul__(self, tr):
        self.validate_transform_for_mul(tr)
        if isinstance(tr, TTransform):
            return TTransform(self.offset + tr.offset, from_cs=tr.systems[0], to_cs=self.systems[1])
        elif isinstance(tr, STTransform):
            return self.as_st() * tr
        elif isinstance(tr, AffineTransform):
            return self.as_affine() * tr
        else:
            return super().__mul__(tr)

    def __rmul__(self, tr):
        tr.validate_transform_for_mul(self)
        if isinstance(tr, STTransform):
            return tr * self.as_st()
        if isinstance(tr, AffineTransform):
            return tr * self.as_affine()
        return super().__rmul__(tr)

    def __repr__(self):
        return f"<TTransform offset={self.offset} at 0x{id(self)}>"


class STTransform(Transform):
    """Transform performing only scale and translate, in that order.

    Input/output dimensionality of this transform may be set by the length of the scale or offset parameters.

    Parameters
    ----------
    scale : array-like
        Scale factors.
    offset : array-like
        Translation distances.
    """

    Linear = True
    Orthogonal = True
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        ArrayParameter("scale", dtype=float, shape=('dims0',), default=1),
        ArrayParameter("offset", dtype=float, shape=('dims0',), default=0),
    ]

    def __init__(self, scale=None, offset=None, dims=None, **kwargs):
        super().__init__(dims, **kwargs, scale=scale, offset=offset)

    def _map(self, coords):
        """Return coordinates mapped by scale and translation.

        Parameters
        ----------
        coords : array-like
            Coordinates to map. The length of the last array dimenstion
            must be equal to the input dimensionality of the transform.

        Returns
        -------
        coords : ndarray
            Mapped coordinates: coords * scale + offset
        """
        return coords * self.scale[None, :] + self.offset[None, :]

    def _imap(self, coords):
        """Return coordinates inverse-mapped by translation and scale.

        Parameters
        ----------
        coords : array-like
            Coordinates to inverse map. The length of the last array dimenstion
            must be equal to the input dimensionality of the transform.

        Returns
        -------
        coords : ndarray
            Mapped coordinates: (coords - offset) / scale
        """
        return (coords - self.offset[None, :]) / self.scale[None, :]

    @property
    def scale(self):
        return self._state["scale"].copy()

    @scale.setter
    def scale(self, s):
        self.set_params(scale=s)

    @property
    def offset(self):
        return self._state["offset"].copy()

    @offset.setter
    def offset(self, t):
        self.set_params(offset=t)

    def translate(self, offset):
        """Change the translation of this transform by the amount given.

        Parameters
        ----------
        offset : array-like
            The values to be added to the current translation of the transform.
        """
        offset = np.asarray(offset)
        self.offset = self.offset + offset

    def zoom(self, zoom, center=None, mapped=True):
        """Update the transform such that its scale factor is changed, but
        the specified center point is left unchanged.

        Parameters
        ----------
        zoom : array-like
            Values to multiply the transform's current scale
            factors.
        center : array-like
            The center point around which the scaling will take place. Defaults to the origin.
        mapped : bool
            Whether *center* is expressed in mapped coordinates (True) or
            unmapped coordinates (False).
        """
        zoom = np.asarray(zoom)
        if center is None:
            center = np.zeros(self.dims[0])
        center = np.asarray(center)
        assert zoom.shape == center.shape == (self.dims[0],)
        scale = self.scale * zoom
        if mapped:
            trans = center - (center - self.offset) * zoom
        else:
            trans = self.scale * (1 - zoom) * center + self.offset
        self.set_params(scale=scale, offset=trans)

    def as_affine(self):
        m = AffineTransform(dims=self.dims, from_cs=self.systems[0], to_cs=self.systems[1])
        m.zoom(self.scale)
        m.translate(self.offset)
        return m

    def as_vispy(self):
        if self.dims != (3, 3):
            raise NotImplementedError("as_vispy is only implemented for 3D transforms")
        from vispy.visuals.transforms import STTransform as VispySTTransform

        return VispySTTransform(scale=self.scale, translate=self.offset)

    @classmethod
    def from_mapping(cls, x0, x1):
        """Create an STTransform from the given mapping

        See `set_mapping` for details.

        Parameters
        ----------
        x0 : array-like
            Start.
        x1 : array-like
            End.

        Returns
        -------
        t : instance of STTransform
            The transform.
        """
        t = cls()
        t.set_mapping(x0, x1)
        return t

    def set_mapping(self, x0, x1):
        """Configure this transform such that it maps points x0 onto x1

        Parameters
        ----------
        x0 : array-like, shape (2, N)
            Two source points
        x1 : array-like, shape (2, N)
            Two destination points

        Examples
        --------
        For example, if we wish to map the corners of a rectangle::

            >>> p1 = [[0, 0], [200, 300]]

        onto a unit cube::

            >>> p2 = [[-1, -1], [1, 1]]

        then we can generate the transform as follows::

            >>> tr = STTransform()
            >>> tr.set_mapping(p1, p2)
            >>> assert tr.map(p1)[:,:2] == p2  # test

        """
        x0 = np.asarray(x0)
        x1 = np.asarray(x1)
        if x0.ndim != 2 or x0.shape[0] != 2 or x1.ndim != 2 or x1.shape[0] != 2:
            raise TypeError("set_mapping requires array inputs of shape " "(2, N).")
        denom = x0[1] - x0[0]
        mask = denom == 0
        denom[mask] = 1.0
        s = (x1[1] - x1[0]) / denom
        s[mask] = 1.0
        t = x1[0] - s * x0[0]
        self.set_params(scale=s, offset=t)

    def __mul__(self, tr):
        self.validate_transform_for_mul(tr)
        if isinstance(tr, STTransform):
            s = self.scale * tr.scale
            t = self.offset + (tr.offset * self.scale)
            return STTransform(scale=s, offset=t, from_cs=tr.systems[0], to_cs=self.systems[1])
        elif isinstance(tr, AffineTransform):
            return self.as_affine() * tr
        else:
            return super().__mul__(tr)

    def __rmul__(self, tr):
        tr.validate_transform_for_mul(self)
        if isinstance(tr, AffineTransform):
            return tr * self.as_affine()
        return super().__rmul__(tr)

    def __repr__(self):
        return f"<STTransform scale={self.scale} offset={self.offset} at 0x{id(self)}>"


class AffineTransform(Transform):
    """Affine transformation class

    Parameters
    ----------
    matrix : array-like | None
        Array to use for the transform. If None, then an identity transform is
        assumed. The shape of the matrix determines the (output, input)
        dimensions of the transform.
    offset : array-like | None
        The translation to apply in this affine transform.
    dims : tuple
        Optionally specifies the (input, output) dimensions of this transform.

    """

    Linear = True
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = False

    parameter_spec = [
        ArrayParameter(
            "matrix",
            dtype=float,
            shape=('dims1', 'dims0'),
            default=lambda shape: np.eye(max(shape))[: shape[1], : shape[0]],
        ),
        ArrayParameter("offset", dtype=float, shape=('dims1',), default=0),
    ]

    def __init__(self, matrix=None, offset=None, dims: Dims = None, **kwargs):
        super().__init__(dims, **kwargs, matrix=matrix, offset=offset)

    def _init_with_no_state(self):
        self._inv_matrix = None
        super()._init_with_no_state()

    def _map(self, coords):
        """Map coordinates

        Parameters
        ----------
        coords : array-like
            Coordinates to map. The length of the last array dimenstion
            must be equal to the input dimensionality of the transform.

        Returns
        -------
        coords : ndarray
            Mapped coordinates: (M * coords) + offset
        """
        if coords.shape[-1] != self.dims[0]:
            raise TypeError(
                "Shape of last axis (%d) is not equal to input dimension of transform (%d)"
                % (coords.shape[-1], self.dims[0])
            )
        return np.dot(self._state["matrix"], coords.T).T + self._state["offset"][None, :]

    def _imap(self, coords):
        """Inverse map coordinates

        Parameters
        ----------
        coords : array-like
            Coordinates to inverse map. The length of the last array dimension
            must be equal to the input dimensionality of the transform.

        Returns
        -------
        coords : ndarray
            Mapped coordinates: M_inv * (coords - offset)
        """
        if coords.shape[-1] != self.dims[1]:
            raise TypeError(
                "Shape of last axis (%d) is not equal to output dimension of transform (%d)"
                % (coords.shape[-1], self.dims[1])
            )
        return np.dot(self.inv_matrix, (coords + self.inv_offset[None, :]).T).T

    @property
    def matrix(self):
        return self._state["matrix"].copy()

    @matrix.setter
    def matrix(self, m):
        self.set_params(matrix=m)

    @property
    def offset(self):
        return self._state["offset"].copy()

    @offset.setter
    def offset(self, o):
        self.set_params(offset=o)

    @property
    def inv_matrix(self):
        if self._inv_matrix is None:
            self._inv_matrix = np.linalg.inv(self._state["matrix"])
        return self._inv_matrix

    @property
    def inv_offset(self):
        return -self._state["offset"]

    def as_affine(self):
        return AffineTransform(
            matrix=self.matrix,
            offset=self.offset,
            from_cs=self.systems[0],
            to_cs=self.systems[1],
        )

    @property
    def full_matrix(self):
        """Return a matrix of shape (N+1, N+1) that contains both self.matrix
        and self.offset::

            [[m11, m21, m31, o1],
             [m21, m22, m32, o2],
             [m31, m32, m33, o3],
             [  0,   0,   0,  1]]

        The full matrix can be multiplied by other similar matrices in order to compose affine
        transforms together.
        """
        m = np.zeros((self.dims[1] + 1, self.dims[0] + 1))
        m[:-1, :-1] = self._state["matrix"]
        m[:-1, -1] = self._state["offset"]
        m[-1, -1] = 1
        return m

    def translate(self, pos):
        """
        Add to the offset.

        The translation is applied *after* the transformations already present.

        Parameters
        ----------
        pos : arrayndarray
            Position to translate by.
        """
        pos = np.asarray(pos)
        self.offset = self.offset + pos

    def zoom(self, scale, center=None):
        """
        Scale the matrix about a given origin.

        The scaling is applied *after* the transformations already present
        in the matrix.

        Parameters
        ----------
        scale : array-like
            Scale factors along x, y and z axes.
        center : array-like or None
            The x, y and z coordinates to scale around. If None,
            (0, 0, 0) will be used.
        """
        if np.isscalar(scale):
            scale = (scale,) * self.dims[0]
        scale_matrix = np.zeros(self.dims[::-1])
        for i in range(min(self.dims)):
            scale_matrix[i, i] = scale[i]

        if center is not None:
            raise NotImplementedError()
        self.matrix = np.dot(scale_matrix, self.matrix)
        self.offset = np.dot(scale_matrix, self.offset)

    def rotate(self, angle, axis=None):
        """
        Rotate the matrix by some angle about a given axis.

        The rotation is applied *after* the transformations already present
        in the matrix.

        Parameters
        ----------
        angle : float
            The angle of rotation in degrees.
        axis : array-like or None
            The x, y and z coordinates of the axis vector to rotate around (only for 3D).
        """
        if self.dims == (2, 2):
            rm = matrices.rotate2d(angle)
        elif self.dims == (3, 3):
            rm = matrices.rotate3d(angle, axis)
        else:
            raise TypeError("Rotation only supported for 2D and 3D affine transforms")
        self.matrix = np.dot(rm, self.matrix)
        self.offset = np.dot(rm, self.offset)

    def set_mapping(self, points1, points2):
        """Set to a transformation matrix that maps points1 onto points2.

        Parameters
        ----------
        points1 : array-like, shape (4, 3)
            Four starting coordinates.
        points2 : array-like, shape (4, 3)
            Four ending coordinates.
        """
        m = matrices.affine_map(points1, points2)
        self.set_params(matrix=m[:, :-1], offset=m[:, -1])

    def __mul__(self, tr):
        self.validate_transform_for_mul(tr)
        if isinstance(tr, AffineTransform):
            m = np.dot(self.full_matrix, tr.full_matrix)
            return AffineTransform(
                matrix=m[:-1, :-1], offset=m[:-1, -1], from_cs=tr.systems[0], to_cs=self.systems[1]
            )
        return tr.__rmul__(self)

    def __eq__(self, tr):
        if not isinstance(tr, AffineTransform):
            # todo: we can assess equality for some others like TTransform and STTransform
            return False
        return np.all(self.full_matrix == tr.full_matrix)

    @classmethod
    def from_qmatrix4x4(cls, matrix, *init_args, **init_kwargs):
        """Create an AffineTransform from a QMatrix4x4."""
        # Extract the matrix data and reshape it to 4x4
        data = matrix.data()
        # QMatrix4x4 data() returns 16 values in column-major order
        matrix_array = np.array(data).reshape(4, 4, order='F')  # Fortran order for column-major

        # Create an AffineTransform from the matrix
        return cls.from_matrix(matrix_array, *init_args, **init_kwargs)

    @classmethod
    def from_matrix(cls, matrix, *init_args, **init_kwargs):
        offset = matrix[:-1, -1]
        matrix = matrix[:-1, :-1]
        return cls(matrix=matrix, offset=offset, *init_args, **init_kwargs)

    def copy(self, from_cs=None, to_cs=None):
        return AffineTransform(
            matrix=self.matrix,
            offset=self.offset,
            from_cs=from_cs or self.systems[0],
            to_cs=to_cs or self.systems[1],
        )


class SRT2DTransform:
    def __init__(self, **kwds):
        raise NotImplementedError()


class SRT3DTransform(Transform):
    """3D transform implemented as 4x4 affine that can always be represented as a combination of 3 matrices: scale * rotate * translate
    This transform has no shear; angles are always preserved.

    Parameters
    ----------
    offset : array-like | None
        Translation distances. Default is (0, 0, 0).
    scale : array-like | None
        Scale factors. Default is (1, 1, 1).
    angle : float | None
        Rotation angle in degrees. Default is 0.
    axis : array-like | None
        Rotation axis. Default is (0, 0, 1).
    """

    Linear = True
    Orthogonal = True
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        ArrayParameter("offset", dtype=float, shape=(3,), default=0),
        ArrayParameter("scale", dtype=float, shape=(3,), default=1),
        ArrayParameter("axis", dtype=float, shape=(3,), default=lambda shape: np.array((0, 0, 1))),
        FloatParameter("angle", default=0.0),
    ]

    def __init__(self, offset=None, scale=None, angle=None, axis=None, init=None, **kwds):
        kwds.setdefault('dims', (3, 3))
        if kwds['dims'] != (3, 3):
            raise ValueError("SRT3DTransform must have dims (3, 3)")
        super().__init__(**kwds, offset=offset, scale=scale, angle=angle, axis=axis)

        # TODO the following looks like untested code
        # if all(p is None for p in (offset, scale, angle, axis)):
        #     if init is not None:
        #         if isinstance(init, SRT3DTransform):
        #             self.set_state(**init._state)
        #         elif isinstance(init, SRT2DTransform):
        #             self.set_state(
        #                 offset=tuple(init._state["offset"]) + (0,),
        #                 scale=tuple(init._state["scale"]) + (1,),
        #                 angle=init._state["angle"],
        #                 axis=(0, 0, 1),
        #             )
        #         elif isinstance(init, AffineTransform):
        #             self.set_from_affine(init)
        #         else:
        #             raise TypeError("Cannot build SRTTransform3D from argument type:", type(init))
        # else:
        #     assert init is None
        #     self.set_params(offset, scale, angle, axis)

    def _init_with_no_state(self):
        super()._init_with_no_state()
        self._affine = None

    @property
    def scale(self):
        return self._state["scale"].copy()

    @scale.setter
    def scale(self, s):
        self.set_params(scale=s)

    @property
    def rotation(self):
        """Return (angle, axis) of rotation"""
        return self._state["angle"], self._state["axis"].copy()

    @rotation.setter
    def rotation(self, angle_axis):
        """Set the transformation rotation to angle (in degrees)"""
        if np.isscalar(angle_axis):
            angle = angle_axis
            axis = (0.0, 0.0, 1.0)
        else:
            angle, axis = angle_axis

        self.set_params(angle=angle, axis=axis)

    @property
    def offset(self):
        return self._state["offset"].copy()

    @offset.setter
    def offset(self, o):
        self.set_params(offset=o)

    def translate(self, offset):
        """Adjust the translation of this transform"""
        self.offset = self._state["offset"] + offset

    def zoom(self, zoom, center=None, mapped=True):
        """Update the transform such that its scale factor is changed, but
        the specified center point is left unchanged.

        Parameters
        ----------
        zoom : array-like
            Values to multiply the transform's current scale
            factors.
        center : array-like
            The center point around which the scaling will take place. Defaults to the origin.
        mapped : bool
            Whether *center* is expressed in mapped coordinates (True) or
            unmapped coordinates (False).
        """
        zoom = np.asarray(zoom)
        if center is None:
            center = np.zeros(self.dims[0])
        center = np.asarray(center)
        assert zoom.shape == center.shape == (self.dims[0],)
        scale = self.scale * zoom
        if mapped:
            trans = center - (center - self.offset) * zoom
        else:
            trans = self.scale * (1 - zoom) * center + self.offset
        self.set_params(scale=scale, offset=trans)

    def rotate(self, angle, axis):
        """Adjust the rotation of this transform"""
        axis = np.asarray(axis)
        origAxis = self._state["axis"]
        if np.all(axis == origAxis):
            self.rotation = self._state["angle"] + angle, axis
        else:
            m = AffineTransform(dims=self.dims)
            m.translate(self._state["offset"])
            m.rotate(self._state["angle"], self._state["axis"])
            m.rotate(angle, axis)
            m.zoom(self._state["scale"])
            self.set_from_affine(m)

    def set_from_affine(self, tr):
        """
        Set this transform based on the elements of *m*
        The input matrix must be affine AND have no shear,
        otherwise the conversion will most likely fail.
        """
        assert tr.dims == (3, 3)

        # scale is vector-length of first three matrix columns
        m = tr.matrix.T
        scale = (m**2).sum(axis=0) ** 0.5

        # see whether there is an inversion
        z = np.cross(m[0], m[1])
        if np.dot(z, m[2]) < 0:
            scale[1] *= -1  ## doesn't really matter which axis we invert

        ## rotation axis is the eigenvector with eigenvalue=1
        r = m / scale[np.newaxis, :]
        try:
            evals, evecs = numpy.linalg.eig(r)
        except Exception:
            print(f"Rotation matrix: {r}")
            print(f"Scale: {scale}")
            print(f"Original matrix: {m}")
            raise
        eigIndex = np.argwhere(np.abs(evals - 1) < 1e-6)
        if len(eigIndex) < 1:
            print(f"eigenvalues: {evals}")
            print(f"eigenvectors: {evecs}")
            print(f"index: {eigIndex}, {evals - 1}")
            raise ValueError("Could not determine rotation axis.")
        axis = evecs[:, eigIndex[0, 0]].real
        axis /= ((axis**2).sum()) ** 0.5

        # trace(r) == 2 cos(angle) + 1, so:
        cos = (r.trace() - 1) * 0.5  # this only gets us abs(angle)

        # The off-diagonal values can be used to correct the angle ambiguity,
        # but we need to figure out which element to use:
        axisInd = np.argmax(np.abs(axis))
        rInd, sign = [((1, 2), -1), ((0, 2), 1), ((0, 1), -1)][axisInd]

        # Then we have r-r.T = sin(angle) * 2 * sign * axis[axisInd];
        # solve for sin(angle)
        sin = (r - r.T)[rInd] / (2.0 * sign * axis[axisInd])

        # finally, we get the complete angle from arctan(sin/cos)
        angle = np.arctan2(sin, cos) * 180 / np.pi
        if angle == 0:
            axis = (0, 0, 1)

        self.set_params(offset=tr.offset, scale=scale, angle=angle, axis=axis)

    @property
    def full_matrix(self):
        return self._get_affine().full_matrix

    def as2D(self):
        """Return an SRT2DTransform representing the x,y portion of this transform (if possible)"""
        return SRT2DTransform(init=self)

    @property
    def params(self):
        return {
            "offset": tuple(self._state["offset"]),
            "scale": tuple(self._state["scale"]),
            "angle": self._state["angle"],
            "axis": tuple(self._state["axis"]),
        }

    def _map(self, arr):
        return self._get_affine()._map(arr)

    def _imap(self, arr):
        return self._get_affine()._imap(arr)

    def set_mapping(self, points1, points2):
        """Set to a transformation that maps points1 onto points2.

        Parameters
        ----------
        points1 : ndarray, shape (N, 3)
            Input coordinates.
        points2 : ndarray, shape (N, 3)
            Output coordinates.
        """
        aff = AffineTransform(dims=(3, 3))
        aff.set_mapping(points1, points2)
        self.set_from_affine(aff)
        return

        params = self.params
        # params_flat = []
        # param_len = {}
        # for k,v in params.items():
        #     if np.isscalar(v):
        #         params_flat.append(v)
        #         param_len[k] = 0
        #     else:
        #         params_flat.extend(v)
        #         param_len[k] = len(v)
        params_flat = (
            list(params["offset"]) + [params["scale"][0]] + list(params["axis"]) + [params["angle"]]
        )
        params_flat = (
            list(params["offset"])
            + list(params["scale"])
            + list(params["axis"])
            + [params["angle"]]
        )
        x0 = np.array(params_flat)

        def unflatten_params(x):
            # return {'offset': x[:3], 'scale': (x[3], x[3], x[3]), 'axis': x[4:7], 'angle': x[7]}
            return {"offset": x[:3], "scale": x[3:6], "axis": x[6:9], "angle": x[9]}
            # params = {}
            # i = 0
            # for k,l in param_len.items():
            #     if l == 0:
            #         params[k] = x[i]
            #         i += 1
            #     else:
            #         params[k] = x[i:i+l]
            #         i += l
            # return params

        def err_func(tr, points1, points2):
            mapped = tr.map(points1)
            err = (mapped - points2).flatten()
            err = (err**2).mean() ** 0.5
            return err

        def flat_err_func(x, points1, points2):
            params = unflatten_params(x)
            tr = SRT3DTransform(**params)
            return err_func(tr, points1, points2)

        # result = scipy.optimize.leastsq(err_func, x0, args=(points1, points2))
        # params = unflatten_params(result[0])
        result = scipy.optimize.minimize(
            flat_err_func,
            x0,
            args=(points1, points2),
            # method=None,
            # method='CG',
            # method=None,
            # method=None,
        )
        params = unflatten_params(result.x)

        self.set_params(**params)
        return err_func(self, points1, points2)

    def as_affine(self):
        affine = AffineTransform(dims=(3, 3), from_cs=self.systems[0], to_cs=self.systems[1])
        affine.zoom(self._state["scale"])
        affine.rotate(self._state["angle"], self._state["axis"])
        affine.translate(self._state["offset"])
        # TODO figure out if this can be generalized to all meta data
        return affine

    def _get_affine(self):
        if getattr(self, "_affine", None) is None:
            self._affine = self.as_affine()
        return self._affine

    def _update(self, source_event=None):
        self._affine = None
        super()._update(source_event=source_event)

    def __repr__(self):
        return (
            f'<SRT3DTransform offset={self._state["offset"]} scale={self._state["scale"]}'
            f' angle={self._state["angle"]} axis={self._state["axis"]} at 0x{id(self)}>'
        )

    def __mul__(self, tr):
        self.validate_transform_for_mul(tr)
        if isinstance(tr, SRT3DTransform):
            return self._get_affine() * tr._get_affine()
        elif isinstance(tr, AffineTransform):
            return self._get_affine() * tr
        else:
            return tr.__rmul__(self)

    @classmethod
    def from_pyqtgraph(cls, pg_transform, *init_args, **init_kwargs):
        """Create an SRT3DTransform from a pyqtgraph Transform3D instance"""
        from pyqtgraph import SRTTransform3D, Transform3D

        if not isinstance(pg_transform, (SRTTransform3D, Transform3D)):
            raise TypeError("Input must be a SRTTransform3D instance")
        offset = pg_transform.getTranslation()
        scale = pg_transform.getScale()
        angle, axis = pg_transform.getRotation()
        angle = -angle  # pyqtgraph uses left-handed rotations
        return cls(*init_args, **init_kwargs, offset=offset, scale=scale, angle=angle, axis=axis)


class PerspectiveTransform(Transform):
    """3D perspective or orthographic matrix transform using homogeneous coordinates.

    Assumes a camera at the origin, looking toward the -Z axis.
    The camera's top points toward +Y, and right points toward +X.

    Points inside the perspective frustum are mapped to the range [-1, +1] along all three axes.
    """

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        TransformParameter("affine", dims=(4, 4), default=lambda: AffineTransform(dims=(4, 4))),
    ]

    def __init__(self, affine=None, **kwds):
        kwds.setdefault('dims', (3, 3))
        if kwds['dims'] != (3, 3):
            raise ValueError("PerspectiveTransform must have dims (3, 3)")
        super().__init__(**kwds, affine=affine)

    @property
    def affine(self):
        return self._state["affine"]

    def _map(self, arr):
        arr4 = np.empty((arr.shape[0], 4), dtype=arr.dtype)
        arr4[:, :3] = arr
        arr4[:, 3] = 1
        out = self.affine.map(arr4)
        return out[:, :3] / out[:, 3:4]

    def as_affine(self):
        # TODO is this correct?
        return self.affine.as_affine()

    @property
    def full_matrix(self):
        # TODO is this correct?
        return self.affine.full_matrix

    def set_ortho(self, left, right, bottom, top, znear, zfar):
        """Set orthographic transform."""
        assert right != left
        assert bottom != top
        assert znear != zfar

        M = np.zeros((4, 4), dtype=np.float32)
        M[0, 0] = +2.0 / (right - left)
        M[3, 0] = -(right + left) / float(right - left)
        M[1, 1] = +2.0 / (top - bottom)
        M[3, 1] = -(top + bottom) / float(top - bottom)
        M[2, 2] = -2.0 / (zfar - znear)
        M[3, 2] = -(zfar + znear) / float(zfar - znear)
        M[3, 3] = 1.0
        self.affine.matrix = M.T

    def set_perspective(self, fovy, aspect, znear, zfar):
        """Set the perspective

        Parameters
        ----------
        fov : float
            Field of view.
        aspect : float
            Aspect ratio.
        near : float
            Near location.
        far : float
            Far location.
        """
        assert znear != zfar
        h = np.tan(fovy * np.pi / 360.0) * znear
        w = h * aspect
        self.set_frustum(-w, w, -h, h, znear, zfar)

    def set_frustum(self, left, right, bottom, top, near, far):  # noqa
        """Set the frustum"""
        M = matrices.frustum(left, right, bottom, top, near, far)
        self.affine.matrix = M.T

    def __eq__(self, tr):
        if not isinstance(tr, PerspectiveTransform):
            return False
        return self.affine == tr.affine


class BilinearTransform(Transform):
    """2D bilinear transform."""

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        ArrayParameter("matrix", dtype=float, shape=(2, 4), default=lambda shape: np.eye(4)[:2]),
        ArrayParameter("inv_matrix", dtype=float, shape=(2, 4), default=lambda shape: np.eye(4)[:2]),
    ]

    def __init__(self, **kwds):
        kwds.setdefault('dims', (2, 2))
        if kwds['dims'] != (2, 2):
            raise ValueError("BilinearTransform must have dims (2, 2)")
        super().__init__(**kwds)

    def set_mapping(self, points1, points2):
        """Set to a transformation matrix that maps points1 onto points2.

        Arguments must be array-like with shape (4, 2).
        """
        # convert inputs to (4, 2) arrays
        points1 = self._prepare_arg_for_mapping(points1)[0]
        points2 = self._prepare_arg_for_mapping(points2)[0]
        assert (
            points1.shape == points2.shape == (4, 2)
        ), f"Input arrays must have shape (4, 2); got ({points1.shape}, {points2.shape})"

        self.set_params(matrix=self._solve_matrix(points1, points2), inv_matrix=self._solve_matrix(points2, points1))

    @staticmethod
    def _solve_matrix(a, b):
        # solve 2 sets of linear equations to determine transformation matrix elements
        c = BilinearTransform._prepare_for_mapping(a)
        matrix = np.zeros((2, 4))
        for i in range(2):
            # solve Ax = B; x is one row of the desired transformation matrix
            matrix[i] = numpy.linalg.solve(c, b[:, i])
        return matrix

    @staticmethod
    def _prepare_for_mapping(points):
        # convert [[x, y], ...] to [[x*y, x, y, 1], ...]
        assert points.ndim >= 2
        assert points.shape[-1] == 2
        out = np.empty(points.shape[:-1] + (4,), dtype=points.dtype)
        out[..., 0] = points[..., 0] * points[..., 1]
        out[..., 1:3] = points
        out[..., 3] = 1
        return out

    @property
    def matrix(self):
        return self._state["matrix"].copy()

    @property
    def inv_matrix(self):
        return self._state["inv_matrix"].copy()

    def _map(self, arr):
        arr4 = BilinearTransform._prepare_for_mapping(arr)
        out = np.dot(arr4, self._state["matrix"].T)
        return out[:, :2]

    def _imap(self, arr):
        arr4 = BilinearTransform._prepare_for_mapping(arr)
        out = np.dot(arr4, self._state["inv_matrix"].T)
        return out[:, :2]


class Homography2DTransform(Transform):
    """2D homography transform."""

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        ArrayParameter("matrix", dtype=float, shape=(3, 3), default=lambda shape: np.eye(3)),
        ArrayParameter("inv_matrix", dtype=float, shape=(3, 3), default=lambda shape: np.eye(3)),
    ]

    def __init__(self, **kwds):
        kwds.setdefault('dims', (2, 2))
        if kwds['dims'] != (2, 2):
            raise ValueError("Homography2DTransform must have dims (2, 2)")
        super().__init__(**kwds)

    def matrix(self):
        return self._state["matrix"].copy()

    @property
    def inv_matrix(self):
        return self._state["inv_matrix"].copy()

    def _map(self, arr):
        arr4 = self._prepare_for_mapping(arr)
        out = np.dot(arr4, self._state["matrix"].T)
        return out[:, :2] / out[:, 2:3]

    def _imap(self, arr):
        arr4 = self._prepare_for_mapping(arr)
        out = np.dot(arr4, self._state["inv_matrix"].T)
        return out[:, :2] / out[:, 2:3]



import pickle
import unittest

import numpy as np
import pytest

from coorx import Point, PointArray, Vector, VectorArray, STTransform
from coorx.systems import CoordinateSystemGraph, CoordinateSystem, get_coordinate_system


def check_point(pt, arr, sys_name):
    """Helper to check Point/PointArray properties."""
    assert isinstance(pt.coordinates, np.ndarray)
    assert pt.ndim == arr.ndim
    assert pt.dtype == arr.dtype
    assert len(pt) == len(arr)
    assert pt.shape == arr.shape  # Shape of the coordinate array
    assert np.allclose(pt.coordinates, arr)  # Use allclose for float comparison
    assert isinstance(pt.system, CoordinateSystem)
    assert pt.system.name == sys_name


def check_vector(vec, p1_arr, p2_arr, sys_name):
    """Helper to check Vector/VectorArray properties."""
    assert isinstance(vec, (Vector, VectorArray))
    assert isinstance(vec.system, CoordinateSystem)
    assert vec.system.name == sys_name
    assert vec.shape == p1_arr.shape or vec.shape == p2_arr.shape
    assert np.allclose(vec.p1.coordinates, p1_arr)
    assert np.allclose(vec.p2.coordinates, p2_arr)
    assert np.allclose(vec.displacement, p2_arr - p1_arr)
    if isinstance(vec, Vector):
        assert isinstance(vec.p1, Point)
        assert isinstance(vec.p2, Point)
    else:
        assert isinstance(vec.p1, PointArray)
        assert isinstance(vec.p2, PointArray)


# Use a fixture for a clean graph per test function if using pytest
# For unittest, manage graph cleanup if necessary (though default graph might be ok)
@pytest.fixture(autouse=True)
def clean_graph():
    # Ensure tests use a clean default graph instance
    CoordinateSystemGraph._graphs = {}
    yield
    CoordinateSystemGraph._graphs = {}


class PointTests(unittest.TestCase):
    def test_init(self):
        # Ensure default graph is clean
        CoordinateSystemGraph.get_graph(create=True)  # Creates 'default' if needed
        # list / tuple / array of int
        d1 = [1, 2, 3]
        for data in [d1, tuple(d1), np.array(d1)]:
            pt = Point(data, "sys3")
            check_point(pt, np.asarray(data), "sys3")

        # list / tuple / array of float
        d2 = [1.0, 2.0, 3.0]
        for data in [d2, tuple(d2), np.array(d2)]:
            pt = Point(data, "sys3")
            check_point(pt, np.asarray(data), "sys3")

        # System ndim check during creation
        sys3 = get_coordinate_system("sys3", ndim=3, create=True)
        assert sys3.ndim == 3
        with self.assertRaisesRegex(TypeError, "System 'sys3' is 3D \\(expected 4D\\)"):
            Point([1, 2, 3, 4], "sys3")  # Now checks against existing system

        # Point must be 1d coordinate array
        with self.assertRaisesRegex(TypeError, "must be 1D"):
            Point([[1, 2, 3]], "sys3")

        # PointArray from list of int - system created implicitly
        data = [[1, 2], [3, 4], [5, 6]]
        sys2 = get_coordinate_system("sys2", ndim=2, create=True)
        arr = PointArray(data, "sys2")
        check_point(arr, np.array(data), "sys2")

        # PointArray from 1d list of Point
        pt1 = Point([1.0, 2.0], "sys2")
        pt2 = Point([3.0, 4.0], "sys2")
        pt3 = Point([5.0, 6.0], "sys2")
        data = [pt1, pt2, pt3]
        arr = PointArray(data, "sys2")
        # np.array(data) extracts coordinates correctly due to __array__
        check_point(arr, np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]), "sys2")

        # PointArray from 2d list of Point
        data = [
            [Point([1.0, 2.0], "sys2"), Point([3.0, 4.0], "sys2")],
            [Point([5.0, 6.0], "sys2"), Point([7.0, 8.0], "sys2")],
        ]
        arr = PointArray(data, "sys2")
        check_point(arr, np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]).reshape(2, 2, 2), "sys2")
        # check that system is carried from Point
        arr = PointArray(data)
        check_point(arr, np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]).reshape(2, 2, 2), "sys2")
        # check conflicting system
        with self.assertRaisesRegex(TypeError, "System wrong_system does not match source coordinates sys2"):
            arr = PointArray(data, "wrong_system")

        # PointArray from 2x2 array of Point
        data = np.empty((2, 2), dtype=object)
        data[:] = [
            [Point([1.0, 2.0], "sys2"), Point([3.0, 4.0], "sys2")],
            [Point([5.0, 6.0], "sys2"), Point([7.0, 8.0], "sys2")],
        ]
        arr = PointArray(data, "sys2")
        check_point(arr, np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]).reshape(2, 2, 2), "sys2")
        # check that system is carried from Point
        arr = PointArray(data)
        check_point(arr, np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]).reshape(2, 2, 2), "sys2")
        # check conflicting system
        with self.assertRaisesRegex(TypeError, "System wrong_system does not match source coordinates sys2"):
            arr = PointArray(data, "wrong_system")

    def test_pickle(self):
        pt = Point([1, 2], "sys2")
        pt2 = pickle.loads(pickle.dumps(pt))
        assert pt == pt2
        assert pt.system is pt2.system  # Should reuse system instance

        arr = PointArray([[1, 2], [3, 4]], "sys2")
        arr2 = pickle.loads(pickle.dumps(arr))
        assert arr == arr2
        assert arr.system is arr2.system

    def test_equality(self):
        p1 = Point([1, 2], "sys2")
        p2 = Point([1, 2], "sys2")
        p3 = Point([1, 3], "sys2")
        p4 = Point([1, 2], "other_sys")
        a1 = PointArray([[1, 2]], "sys2")

        assert p1 == p2
        assert p1 != p3
        assert p1 != p4  # Different system
        assert p1 != a1  # Different type
        assert p1 != [1, 2]  # Different type


# ==== Vector Tests ====


class VectorTests(unittest.TestCase):
    def test_vector_init(self):
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        v = Vector(p1, p2)
        check_vector(v, np.array([1, 2]), np.array([4, 6]), "cartesian")
        assert v.shape == (2,)  # Shape of the coordinate array for a single point
        assert v.ndim == 1  # ndim of the coordinate array
        assert len(v) == 2  # Length of the coordinate array

        # Mismatched systems
        p3 = Point([0, 0], "polar")
        with self.assertRaisesRegex(ValueError, "must share the same coordinate system"):
            Vector(p1, p3)

        # Wrong types
        with self.assertRaisesRegex(TypeError, "must be Point instances"):
            Vector(p1, [4, 6])
        with self.assertRaisesRegex(TypeError, "must be Point instances"):
            Vector(PointArray([[1, 2]], "cartesian"), p2)

    def test_vector_array_init(self):
        pa1 = PointArray([[1, 2], [0, 0]], "cartesian")
        pa2 = PointArray([[4, 6], [1, 1]], "cartesian")
        va = VectorArray(pa1, pa2)
        check_vector(va, np.array([[1, 2], [0, 0]]), np.array([[4, 6], [1, 1]]), "cartesian")
        assert va.shape == (2, 2)  # Shape of the point array
        assert va.ndim == 2  # ndim of the point array structure
        assert len(va) == 2  # Number of vectors

        # Mismatched systems
        pa3 = PointArray([[0, 0], [0, 0]], "polar")
        with self.assertRaisesRegex(ValueError, "must share the same coordinate system"):
            VectorArray(pa1, pa3)

        # Mismatched shapes
        pa4 = PointArray([[0, 0]], "cartesian")
        with self.assertRaisesRegex(ValueError, "must have the same structural shape"):
            VectorArray(pa1, pa4)

        # Wrong types
        with self.assertRaisesRegex(TypeError, "must be PointArray or Point instances"):
            VectorArray(pa1, [[4, 6], [1, 1]])

        # Can init with Points (they are PointArrays)
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        va_from_points = VectorArray(p1, p2)  # Technically allowed, but Vector is preferred
        check_vector(va_from_points, np.array([1, 2]), np.array([4, 6]), "cartesian")

    def test_vector_array_init_dimension_validation(self):
        """Test that VectorArray validates dimension compatibility between endpoints."""
        # Mismatched coordinate dimensions should fail when systems differ
        pa_2d = PointArray([[1, 2], [3, 4]], "cartesian")  # 2D coordinates
        pa_3d = PointArray([[1, 2, 3], [4, 5, 6]], "cartesian3d")  # 3D coordinates
        
        with self.assertRaisesRegex(ValueError, "must share the same coordinate system"):
            VectorArray(pa_2d, pa_3d)  # Different systems will catch this first
        
        # Test what happens when trying to create PointArrays with mismatched dimensions
        # in the same system. This should fail at PointArray creation time.
        try:
            # This should fail - cartesian is 2D, can't give it 3D coordinates  
            invalid_pa = PointArray([[1, 2, 3]], "cartesian")
            assert False, "Should have failed with dimension mismatch"
        except TypeError as e:
            assert "is 2D (expected 3D)" in str(e)
        
        # Valid case: same dimensions in same system should work
        pa1 = PointArray([[1, 2], [3, 4]], "cartesian")  
        pa2 = PointArray([[5, 6], [7, 8]], "cartesian")  
        va = VectorArray(pa1, pa2)
        assert va.shape == (2, 2)

    def test_vector_mixed_init_validations(self):
        pa_2d = PointArray([[1, 2], [3, 4]], "cartesian")
        p_2d = Point([5, 6], "cartesian")
        pa_3d = PointArray([[1, 2, 3], [4, 5, 6]], "cartesian3d")
        p_3d = Point([7, 8, 9], "cartesian3d")
        # Valid mixed init: Point with PointArray
        va1 = VectorArray(p_2d, pa_2d)
        check_vector(va1, np.array([5, 6]), pa_2d.coordinates, "cartesian")
        assert va1.shape == (2, 2)  # Should broadcast to match PointArray length
        # Valid mixed init: PointArray with Point
        va2 = VectorArray(pa_2d, p_2d)
        check_vector(va2, pa_2d.coordinates, np.array([5, 6]), "cartesian")
        assert va2.shape == (2, 2)  # Should broadcast to match PointArray length
        # Valid mixed init: Point with Point
        va3 = VectorArray(p_3d, pa_3d)
        check_vector(va3, np.array([7, 8, 9]), pa_3d.coordinates, "cartesian3d")
        assert va3.shape == (2, 3)  # Should broadcast to match PointArray length
        # Valid mixed init: PointArray with Point
        va4 = VectorArray(pa_3d, p_3d)
        check_vector(va4, pa_3d.coordinates, np.array([7, 8, 9]), "cartesian3d")
        assert va4.shape == (2, 3)  # Should broadcast to match PointArray length
        with self.assertRaisesRegex(ValueError, "must share the same coordinate system"):
            # Invalid mixed init: mismatched systems
            VectorArray(p_2d, pa_3d)
        with self.assertRaisesRegex(ValueError, "must share the same coordinate system"):
            # Invalid mixed init: mismatched systems
            VectorArray(pa_2d, p_3d)

    def test_vector_array_init_length_validation(self):
        """Test that VectorArray validates length compatibility when both endpoints are PointArrays."""
        # PointArrays with different lengths should fail
        pa_short = PointArray([[1, 2]], "cartesian")  # Length 1
        pa_long = PointArray([[3, 4], [5, 6], [7, 8]], "cartesian")  # Length 3
        
        with self.assertRaisesRegex(ValueError, "must have the same structural shape"):
            VectorArray(pa_short, pa_long)
        
        # Point with PointArray should work (broadcasting)
        p = Point([1, 2], "cartesian")
        pa = PointArray([[3, 4], [5, 6]], "cartesian")
        va = VectorArray(p, pa)
        assert va.shape == (2, 2)  # Should broadcast to match PointArray length
        
        # PointArray with Point should also work
        va2 = VectorArray(pa, p)
        assert va2.shape == (2, 2)
        
        # Same length PointArrays should work
        pa1 = PointArray([[1, 2], [3, 4]], "cartesian")
        pa2 = PointArray([[5, 6], [7, 8]], "cartesian")
        va3 = VectorArray(pa1, pa2)
        assert va3.shape == (2, 2)
        
        # Test verification that the validation requested is already working
        # Verify: PointArrays must have same coordinate dimensionality (enforced by system check)
        pa_2d = PointArray([[1, 2]], "cartesian")  # 2D
        pa_3d = PointArray([[1, 2, 3]], "cartesian3d")  # 3D
        with self.assertRaisesRegex(ValueError, "must share the same coordinate system"):
            VectorArray(pa_2d, pa_3d)  # Different coordinate systems (2D vs 3D)

    def test_point_subtraction(self):
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        p3 = Point([0, 0], "polar")

        # Point - Point
        v = p2 - p1
        assert isinstance(v, Vector)
        check_vector(v, p1.coordinates, p2.coordinates, "cartesian")

        # Mismatched systems
        with self.assertRaisesRegex(ValueError, "does not match"):
            p2 - p3

        # PointArray - PointArray
        pa1 = PointArray([[1, 2], [0, 0]], "cartesian")
        pa2 = PointArray([[4, 6], [1, 1]], "cartesian")
        va = pa2 - pa1
        assert isinstance(va, VectorArray)
        assert not isinstance(va, Vector)  # Ensure it's not the specific subclass
        check_vector(va, pa1.coordinates, pa2.coordinates, "cartesian")

        # PointArray - Point (broadcasts conceptually) -> VectorArray
        va2 = pa1 - p1
        assert isinstance(va2, VectorArray)
        # Expected p1 coords broadcasted to match pa1 shape for subtraction endpoint
        expected_p1_arr = np.array([[1, 2], [1, 2]])
        check_vector(va2, expected_p1_arr, pa1.coordinates, "cartesian")

        # Point - PointArray -> VectorArray
        va3 = p2 - pa1
        assert isinstance(va3, VectorArray)
        expected_p2_arr = np.array([[4, 6], [4, 6]])
        check_vector(va3, pa1.coordinates, expected_p2_arr, "cartesian")

    def test_vector_addition(self):
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        p3 = Point([0, 5], "cartesian")
        p4 = Point([1, 0], "cartesian")
        p_polar = Point([0, 0], "polar")

        v1 = Vector(p1, p2)  # Displacement [3, 4]
        v2 = Vector(p3, p4)  # Displacement [1, -5]

        # Vector + Vector
        v_sum = v1 + v2
        assert isinstance(v_sum, Vector)
        # Expected displacement: [3, 4] + [1, -5] = [4, -1]
        # Expected endpoints: starts at v1.p1, ends at v1.p1 + combined_disp
        # p1 = [1, 2], combined_disp = [4, -1] -> p2_new = [5, 1]
        check_vector(v_sum, p1.coordinates, np.array([5, 1]), "cartesian")
        assert np.allclose(v_sum.displacement, [4, -1])

        # Mismatched systems
        v_polar = Vector(p_polar, p_polar)
        with self.assertRaisesRegex(ValueError, "does not match"):
            v1 + v_polar

        # VectorArray + VectorArray
        pa1 = PointArray([[1, 2], [0, 0]], "cartesian")
        pa2 = PointArray([[4, 6], [1, 1]], "cartesian")  # Disp: [[3, 4], [1, 1]]
        pa3 = PointArray([[0, 5], [-1, 0]], "cartesian")
        pa4 = PointArray([[1, 0], [-1, 2]], "cartesian")  # Disp: [[1, -5], [0, 2]]
        va1 = VectorArray(pa1, pa2)
        va2 = VectorArray(pa3, pa4)

        va_sum = va1 + va2
        assert isinstance(va_sum, VectorArray)
        # Expected displacement: [[3, 4], [1, 1]] + [[1, -5], [0, 2]] = [[4, -1], [1, 3]]
        # Expected endpoints: starts at va1.p1, ends at va1.p1 + combined_disp
        # p1 = [[1, 2], [0, 0]], combined_disp = [[4, -1], [1, 3]] -> p2_new = [[5, 1], [1, 3]]
        check_vector(va_sum, pa1.coordinates, np.array([[5, 1], [1, 3]]), "cartesian")
        assert np.allclose(va_sum.displacement, [[4, -1], [1, 3]])

        # Vector + VectorArray
        vsum_va = v1 + va1
        assert isinstance(vsum_va, VectorArray)
        # Disp v1: [3, 4]. Disp va1: [[3, 4], [1, 1]]
        # Combined: [[6, 8], [4, 5]] (broadcasting v1)
        # Starts at v1.p1 broadcasted: [[1, 2], [1, 2]]
        # Ends at start + combined: [[7, 10], [5, 7]]
        check_vector(vsum_va, np.array([[1, 2], [1, 2]]), np.array([[7, 10], [5, 7]]), "cartesian")

        # VectorArray + Vector
        vasum_v = va1 + v1
        assert isinstance(vasum_v, VectorArray)
        # Combined: [[6, 8], [4, 5]]
        # Starts at va1.p1: [[1, 2], [0, 0]]
        # Ends at start + combined: [[7, 10], [4, 5]]
        check_vector(vasum_v, np.array([[1, 2], [0, 0]]), np.array([[7, 10], [4, 5]]), "cartesian")

    def test_point_vector_addition(self):
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        v1 = Vector(p1, p2)  # Displacement [3, 4]
        p_other = Point([-1, -1], "cartesian")
        p_polar = Point([0, 0], "polar")

        # Point + Vector
        p_new = p_other + v1
        assert isinstance(p_new, Point)
        # Expected coords: [-1, -1] + [3, 4] = [2, 3]
        check_point(p_new, np.array([2, 3]), "cartesian")

        # Vector + Point (using __radd__)
        p_new_r = v1 + p_other
        assert isinstance(p_new_r, Point)
        check_point(p_new_r, np.array([2, 3]), "cartesian")

        # Mismatched systems
        with self.assertRaisesRegex(ValueError, "does not match"):
            p_polar + v1
        with self.assertRaisesRegex(ValueError, "does not match"):
            v1 + p_polar

        # PointArray + VectorArray
        pa1 = PointArray([[1, 2], [0, 0]], "cartesian")
        pa2 = PointArray([[4, 6], [1, 1]], "cartesian")  # Disp: [[3, 4], [1, 1]]
        va1 = VectorArray(pa1, pa2)
        pa_other = PointArray([[10, 20], [30, 40]], "cartesian")

        pa_new = pa_other + va1
        assert isinstance(pa_new, PointArray)
        # Expected coords: [[10, 20], [30, 40]] + [[3, 4], [1, 1]] = [[13, 24], [31, 41]]
        check_point(pa_new, np.array([[13, 24], [31, 41]]), "cartesian")

        # VectorArray + PointArray
        pa_new_r = va1 + pa_other
        assert isinstance(pa_new_r, PointArray)
        check_point(pa_new_r, np.array([[13, 24], [31, 41]]), "cartesian")

        # PointArray + Vector
        pa_new_v = pa_other + v1  # v1 disp [3, 4] broadcasted
        assert isinstance(pa_new_v, PointArray)
        # Expected: [[10, 20], [30, 40]] + [[3, 4], [3, 4]] = [[13, 24], [33, 44]]
        check_point(pa_new_v, np.array([[13, 24], [33, 44]]), "cartesian")

        # Vector + PointArray
        pa_new_v_r = v1 + pa_other
        assert isinstance(pa_new_v_r, PointArray)
        check_point(pa_new_v_r, np.array([[13, 24], [33, 44]]), "cartesian")

        # Point + VectorArray
        p_new_va = p1 + va1  # p1 [1, 2] broadcasted
        assert isinstance(p_new_va, PointArray)
        # Expected: [[1, 2], [1, 2]] + [[3, 4], [1, 1]] = [[4, 6], [2, 3]]
        check_point(p_new_va, np.array([[4, 6], [2, 3]]), "cartesian")

        # VectorArray + Point
        p_new_va_r = va1 + p1
        assert isinstance(p_new_va_r, PointArray)
        check_point(p_new_va_r, np.array([[4, 6], [2, 3]]), "cartesian")

    def test_point_vector_subtraction(self):
        # Setup common points and vectors
        p_start = Point([10, 20], "cartesian")
        v_disp_p1 = Point([1, 2], "cartesian")
        v_disp_p2 = Point([4, 6], "cartesian")  # Displacement [3, 4]
        vector = Vector(v_disp_p1, v_disp_p2)

        pa_start = PointArray([[10, 20], [30, 40]], "cartesian")
        va_disp_p1 = PointArray([[1, 2], [0, 0]], "cartesian")
        va_disp_p2 = PointArray([[4, 6], [1, 1]], "cartesian")  # Displacements [[3,4], [1,1]]
        vector_array = VectorArray(va_disp_p1, va_disp_p2)

        p_polar = Point([0, 0], "polar")
        v_polar_disp_p1 = Point([0, 0], "polar")
        v_polar_disp_p2 = Point([1, 1], "polar")
        vector_polar = Vector(v_polar_disp_p1, v_polar_disp_p2)

        # Case 1: Point - Vector
        result_pv = p_start - vector
        assert isinstance(result_pv, Point)
        # Expected coords: [10, 20] - [3, 4] = [7, 16]
        check_point(result_pv, np.array([7, 16]), "cartesian")

        # Case 2: Point - VectorArray
        result_pva = p_start - vector_array
        assert isinstance(result_pva, PointArray)
        # Expected coords: [[10, 20], [10, 20]] - [[3, 4], [1, 1]] = [[7, 16], [9, 19]]
        check_point(result_pva, np.array([[7, 16], [9, 19]]), "cartesian")

        # Case 3: PointArray - Vector
        result_pav = pa_start - vector
        assert isinstance(result_pav, PointArray)
        # Expected coords: [[10, 20], [30, 40]] - [[3, 4], [3, 4]] = [[7, 16], [27, 36]]
        check_point(result_pav, np.array([[7, 16], [27, 36]]), "cartesian")

        # Case 4: PointArray - VectorArray
        result_pava = pa_start - vector_array
        assert isinstance(result_pava, PointArray)
        # Expected coords: [[10, 20], [30, 40]] - [[3, 4], [1, 1]] = [[7, 16], [29, 39]]
        check_point(result_pava, np.array([[7, 16], [29, 39]]), "cartesian")

        # Error handling: Mismatched systems
        with self.assertRaisesRegex(ValueError, "does not match this PointArray's system"):
            p_start - vector_polar
        with self.assertRaisesRegex(ValueError, "does not match this PointArray's system"):
            pa_start - vector_polar

        # Error handling: Type errors (subtracting non-vector from point)
        # This should be caught by PointArray.__sub__ if the operand is not PointArray or VectorArray
        with self.assertRaisesRegex(TypeError, "Unsupported operand type"):  # Or similar, depending on implementation
            p_start - np.array([1, 2])
        with self.assertRaisesRegex(TypeError, "Unsupported operand type"):
            pa_start - np.array([[1, 2]])

    def test_vector_pickle(self):
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        v = Vector(p1, p2)
        v2 = pickle.loads(pickle.dumps(v))
        assert v == v2
        assert v.system is v2.system  # System should be reused

        pa1 = PointArray([[1, 2], [0, 0]], "cartesian")
        pa2 = PointArray([[4, 6], [1, 1]], "cartesian")
        va = VectorArray(pa1, pa2)
        va2 = pickle.loads(pickle.dumps(va))
        assert va == va2
        assert va.system is va2.system

    def test_vector_equality(self):
        p1 = Point([1, 2], "sys")
        p2 = Point([3, 4], "sys")
        p3 = Point([1, 2], "sys")  # Same coords as p1
        p4 = Point([3, 5], "sys")  # Different coords from p2
        p5 = Point([1, 2], "other_sys")
        p6 = Point([3, 4], "other_sys")

        v1 = Vector(p1, p2)
        v2 = Vector(p1, p2)  # Identical
        v3 = Vector(p3, p2)  # Same start point value, different object
        v4 = Vector(p1, p4)  # Different end point
        v5 = Vector(p5, p6)  # Different system

        va1 = VectorArray(PointArray([p1.coordinates], "sys"), PointArray([p2.coordinates], "sys"))

        assert v1 == v2
        assert v1 == v3  # Equality based on endpoint values
        assert v1 != v4
        assert v1 != v5  # Different system via endpoints
        assert v1 != va1  # Different type
        assert v1 != [2, 2]  # Different type (displacement)

    def test_vector_simplified_init_only_points(self):
        """Test that VectorArray only accepts two Point/PointArray instances, not arrays."""
        # VectorArray should only accept Point/PointArray instances
        p1 = Point([1, 2], "cartesian")
        p2 = Point([4, 6], "cartesian")
        v = Vector(p1, p2)
        assert isinstance(v, Vector)
        check_vector(v, np.array([1, 2]), np.array([4, 6]), "cartesian")

        # VectorArray with PointArrays should work
        pa1 = PointArray([[1, 2], [3, 4]], "cartesian")
        pa2 = PointArray([[5, 6], [7, 8]], "cartesian")
        va = VectorArray(pa1, pa2)
        assert isinstance(va, VectorArray)
        check_vector(va, np.array([[1, 2], [3, 4]]), np.array([[5, 6], [7, 8]]), "cartesian")

        # Array + system initialization should no longer be supported
        disp = np.array([3, 4])
        with self.assertRaisesRegex(TypeError, "must be Point instances"):
            Vector(disp, "cartesian")
        
        multi_disp = np.array([[1, 2], [3, 4]])
        with self.assertRaisesRegex(TypeError, "must be PointArray or Point instances"):
            VectorArray(multi_disp, "cartesian")

    def test_vector_mapped_to(self):
        # Setup a graph with a transform
        graph = CoordinateSystemGraph.get_graph("test_graph", create=True)
        graph.add_system("sys_a", ndim=2)
        graph.add_system("sys_b", ndim=2)

        # A transform that scales by 2 and translates by (10, 20)
        tr = STTransform(scale=[2, 2], offset=[10, 20])
        graph.add_transform(tr, "sys_a", "sys_b")

        # 1. Test Vector.mapped_to
        p1 = Point([1, 2], "sys_a", graph="test_graph")
        p2 = Point([4, 6], "sys_a", graph="test_graph")
        v = Vector(p1, p2)

        # The displacement in sys_a is [3, 4]
        assert np.allclose(v.displacement, [3, 4])

        # Map the vector to sys_b
        v_mapped = v.mapped_to("sys_b")

        # Check the mapped vector
        assert isinstance(v_mapped, Vector)
        assert v_mapped.system.name == "sys_b"

        p1_mapped_coords = np.array([1, 2]) * 2 + [10, 20]  # [12, 24]
        p2_mapped_coords = np.array([4, 6]) * 2 + [10, 20]  # [18, 32]

        check_vector(v_mapped, p1_mapped_coords, p2_mapped_coords, "sys_b")
        assert np.allclose(v_mapped.displacement, [6, 8])

        # 2. Test VectorArray.mapped_to
        pa1 = PointArray([[1, 2], [0, 0]], "sys_a", graph="test_graph")
        pa2 = PointArray([[4, 6], [1, 1]], "sys_a", graph="test_graph")
        va = VectorArray(pa1, pa2)

        # Displacements in sys_a are [[3, 4], [1, 1]]
        assert np.allclose(va.displacement, [[3, 4], [1, 1]])

        # Map the vector array to sys_b
        va_mapped = va.mapped_to("sys_b")

        assert isinstance(va_mapped, VectorArray)
        assert not isinstance(va_mapped, Vector)
        assert va_mapped.system.name == "sys_b"

        # Map endpoints for VectorArray
        pa1_mapped_coords = np.array([[1, 2], [0, 0]]) * 2 + [10, 20]  # [[12, 24], [10, 20]]
        pa2_mapped_coords = np.array([[4, 6], [1, 1]]) * 2 + [10, 20]  # [[18, 32], [12, 22]]

        check_vector(va_mapped, pa1_mapped_coords, pa2_mapped_coords, "sys_b")
        assert np.allclose(va_mapped.displacement, [[6, 8], [2, 2]])



import numpy as np


class Parameter:
    def __init__(self, name):
        self.name = name

    def validate(self, value, old_value, dims):
        """Return the validated value for this parameter and whether it has changed.
        Raise an exception if the value is invalid.
        """
        return value, value != old_value

    def check_dims(self, unvalidated_value, dims) -> bool | None:
        """Return True if using *unvalidated_value* for this parameter is compatible with *dims*.
        Return None if the parameter does not depend on dims.
        """
        return None

    def infer_dims(self, unvalidated_value) -> tuple[float | None, float | None]:
        """Return the dims that would be required to use *value* for this parameter.
        Return None for either dim if the parameter does not constrain it.
        """
        return None, None


class FloatParameter(Parameter):
    def __init__(self, name, default=None):
        super().__init__(name)
        self.default = default

    def validate(self, value, old_value, dims):
        if value is None and self.default is not None:
            value = self.default
        try:
            value = float(value)
        except Exception as e:
            raise ValueError(f"Parameter '{self.name}' must be a float") from e
        return value, value != old_value


class TupleParameter(Parameter):
    def __init__(self, name, length=None, dtype=None, default=None):
        super().__init__(name)
        if length is not None and length not in ('dims0', 'dims1') and not isinstance(length, int):
            raise ValueError("length must be None, integer, or 'dims0'/'dims1'")
        self.length = length
        self.dtype = dtype
        self.default = default

    def validate(self, value, old_value, dims):
        # make sure value is tuple-like
        if value is None and self.default is not None:
            value = self.default(dims)
        try:
            value = tuple(value)
        except Exception as e:
            raise TypeError(f"Parameter '{self.name}' must be tuple-like") from e

        # check length
        length = self.length
        if length in ('dims0', 'dims1'):
            length = {'dims0': dims[0], 'dims1': dims[1]}[length]
        if length is not None and len(value) != length:
            raise ValueError(f"Parameter '{self.name}' must have length {length}")

        # convert elements to dtype
        if self.dtype is not None:
            try:
                value = tuple(self.dtype(v) for v in value)
            except Exception as e:
                raise TypeError(
                    f"Elements of parameter '{self.name}' must be of type {self.dtype.__name__}"
                ) from e

        return value, value != old_value

    def check_dims(self, unvalidated_value, dims):
        if unvalidated_value is None:
            return None
        if isinstance(self.length, str):
            expected_length = dims[0] if self.length == "dims0" else dims[1]
            return len(unvalidated_value) == expected_length
        return None

    def infer_dims(self, unvalidated_value):
        if unvalidated_value is None:
            return None, None
        if isinstance(self.length, str):
            return (
                (len(unvalidated_value), None)
                if self.length == "dims0"
                else (None, len(unvalidated_value))
            )
        return None, None


class ArrayParameter(Parameter):
    def __init__(self, name, dtype=float, shape=None, default=None):
        super().__init__(name)
        if not isinstance(shape, tuple) and shape is not None:
            raise ValueError("shape must be a tuple or None")
        if shape is not None:
            for x in shape:
                if not (x is None or isinstance(x, int) or x in ("dims0", "dims1")):
                    raise ValueError("shape elements must be None, integer, or 'dims0'/'dims1'")
        self.dtype = dtype
        self.shape = shape
        self.default = default

    def validate(self, value, old_value, dims):
        # determine shape constraints
        shape = self.shape
        if shape is not None:
            str_dims = {"dims0": dims[0], "dims1": dims[1]}
            shape = tuple([str_dims.get(dim, dim) for dim in shape])

        # handle default
        if value is None:
            if self.default is None:
                raise ValueError(f"Parameter '{self.name}' cannot be None")
            if shape is None or None in shape:
                raise ValueError(
                    f"Parameter '{self.name}' has undefined shape, cannot use callable default"
                )
            if callable(self.default):
                value = self.default(shape)
            elif self.default is not None:
                value = np.ones(shape, dtype=self.dtype) * self.default

        # check array / dtype
        try:
            value = np.asarray(value, dtype=self.dtype)
        except Exception as e:
            dtype_err = ' of type ' + self.dtype.__name__ if self.dtype is not None else ''
            raise TypeError(f"Parameter '{self.name}' must be an array{dtype_err}") from e

        # check shape
        if shape is not None:
            assert (
                len(shape) == value.ndim
            ), f"Parameter '{self.name}' must have ndim {len(shape)}, got {value.ndim}"
            for i, dim in enumerate(shape):
                if dim is not None and value.shape[i] != dim:
                    raise ValueError(
                        f"Parameter '{self.name}' must have shape[{i}]=={shape[i]}, got {value.shape[i]}"
                    )

        return value, not np.array_equal(value, old_value)

    def check_dims(self, unvalidated_value, dims):
        if unvalidated_value is None:
            return None
        unvalidated_value = np.asarray(unvalidated_value)
        if self.shape is not None:
            str_dims = {"dims0": dims[0], "dims1": dims[1]}
            for i, dim in enumerate(self.shape):
                expected_dim = str_dims.get(dim, dim)
                if expected_dim is not None and unvalidated_value.shape[i] != expected_dim:
                    return False
        return None

    def infer_dims(self, unvalidated_value):
        dims0 = None
        dims1 = None
        if unvalidated_value is None:
            return dims0, dims1
        unvalidated_value = np.asarray(unvalidated_value)
        if self.shape is not None:
            for dim in self.shape:
                if dim == "dims0":
                    dims0 = unvalidated_value.shape[self.shape.index(dim)]
                elif dim == "dims1":
                    dims1 = unvalidated_value.shape[self.shape.index(dim)]
        return dims0, dims1


class TransformParameter(Parameter):
    def __init__(self, name, dims=None, default=None):
        super().__init__(name)
        self.dims = dims
        self.default = default

    def validate(self, value, old_value, dims):
        from . import Transform, create_transform

        if isinstance(value, dict):
            try:
                value = create_transform(**value)
            except Exception as e:
                raise ValueError(
                    f"Transform for parameter {self.name} could not be created from {value}"
                ) from e
        if value is None and callable(self.default):
            value = self.default()
        if not isinstance(value, Transform):
            raise TypeError(
                f"Parameter '{self.name}' must be a Transform instance, not {type(value).__name__}"
            )
        if self.dims is not None and value.dims != self.dims:
            raise ValueError(
                f"Parameter '{self.name}' must have dims {self.dims}, got {value.dims}"
            )
        changed = value != old_value
        return value, changed


class TransformListParameter(Parameter):
    def __init__(self, name):
        super().__init__(name)

    def validate(self, new_value, old_value, dims):
        raise NotImplementedError("this isn't being used yet")
        from . import create_transform, Transform
        if new_value is None:
            new_value = []
        new_value = [t if isinstance(t, Transform) else create_transform(**t) for t in new_value]
        changed = new_value != old_value
        return new_value, changed



from __future__ import annotations

import numpy as np

from .systems import CoordinateSystem, CoordinateSystemGraph, get_coordinate_system
from ._types import StrOrNone, CoordSysOrStr


class PointArray:
    """Represents an N-dimensional array of points.

    Although the array is instantiated with the coordinates of these points in a particular
    coordinate system, one may request the coordinates mapped to any other system.

    Parameters
    ----------
    coordinates : array-like
        Array or list of coordinates. These may be numerical or Point instances.
    system : str | System | None
        The coordinate system to which th points in *coordinates* belong.
    graph : str | CoordinateSystemGraph
        If *system* is a string, then the coordinate system is looked up from this graph.
    """

    def __init__(self, coordinates, system: CoordSysOrStr = None, graph: StrOrNone = None):
        coord_arr, source_system = self._interpret_input(coordinates)

        if not np.issubdtype(coord_arr.dtype, np.number):
            raise TypeError(f"Coordinates must be numerical or Point instances, not {coord_arr.dtype}")
        self._coordinates = coord_arr

        # get the requested coordinate system
        self.system = None
        if system is not None:
            self.set_system(system, graph)

        # if the input came with a coordinate system, use that and
        # verify that it does not conflict with the previously specified system
        if source_system is not None:
            if self.system not in (None, source_system):
                raise TypeError(f"System {system} does not match source coordinates {source_system}")
            self.system = source_system

        if self.system is not None and self.shape[-1] != self.system.ndim:
            raise TypeError(f"System ndim is {self.system.ndim}, but coordinate data is {self.shape[-1]}D")

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def shape(self):
        return self.coordinates.shape

    @property
    def ndim(self):
        return self.coordinates.ndim

    @property
    def dtype(self):
        return self.coordinates.dtype

    def __len__(self):
        return len(self.coordinates)

    def __getitem__(self, index):
        return self.coordinates[index]

    def __iter__(self):
        yield from self.coordinates

    def __add__(self, b):
        if isinstance(b, (Vector, VectorArray)):
            self._check_vector_operand(b)
            new_coords = self.coordinates + b.displacement
            if isinstance(self, Point) and isinstance(b, Vector):
                return Point(new_coords, system=self.system)
            else:
                return PointArray(new_coords, system=self.system)
        else:
            self._check_point_operand(b)
            # This case should likely not return raw coordinates, but maybe another PointArray?
            # For now, keeping original behavior for PointArray + PointArray.
            # Consider if PointArray + PointArray should be allowed or raise TypeError.
            return self.coordinates + b.coordinates

    def __sub__(self, b: PointArray | VectorArray) -> VectorArray | PointArray:
        if isinstance(b, PointArray):
            self._check_point_operand(b)
            return VectorArray(b, self)
        elif isinstance(b, VectorArray):
            self._check_vector_operand(b)
            new_coords = self.coordinates - b.displacement
            return PointArray(new_coords, system=self.system)
        raise TypeError("Unsupported operand type")

    def mapped_through(self, cs_list) -> PointArray:
        chain = self.system.graph.transform_chain([self.system] + cs_list)
        return chain.map(self)

    def mapped_to(self, system) -> "PointArray":
        path = self.system.graph.transform_path(self.system, system)
        chain = self.system.graph.transform_chain(path)
        return chain.map(self)

    def set_system(self, system, graph=None):
        self.system = get_coordinate_system(system, graph=graph, ndim=self.shape[-1], create=True)

    def _coorx_transform(self, tr):
        if tr.systems[0] is not self.system:
            raise TypeError(
                f"The transform {tr} maps from system {tr.systems[0]}, but this PointArray is defined in {self.system}"
            )
        # Map the raw coordinates
        mapped_coords = tr.map(self.coordinates)
        output_system = tr.systems[1]
        if isinstance(self, Point):
            return Point(mapped_coords, system=output_system)
        else:
            return PointArray(mapped_coords, system=output_system)

    def __array__(self, dtype=None):
        # Ensure subclasses like Point also work correctly with np.asarray
        coords = self.coordinates
        if dtype is None:
            return coords
        else:
            return coords.astype(dtype, copy=False)

    def __repr__(self):
        return f"<{type(self).__name__} {self.shape} in {self.system.name}>"

    def __getstate__(self):
        state = self.__dict__.copy()
        state["system"] = None if self.system is None else (self.system.name, self.system.graph.name)
        return state

    def __setstate__(self, state):
        sys, graph = state.pop("system")
        self.__dict__.update(state)
        self.set_system(sys, graph)

    def __eq__(self, b):
        if type(b) is not type(self):
            return False
        if self.coordinates.shape != b.coordinates.shape:
            return False
        if not np.all(self.coordinates == b.coordinates):
            return False
        if self.system is not b.system:
            return False
        return True

    def _interpret_input(self, coordinates):
        """Interpret initial coordinates argument, return a numerical array of coordinates
        and a coordinate system if it was present in the input.
        """
        # convert coordinates to array
        coord_arr = np.asarray(coordinates)
        if coord_arr.size == 0:
            raise ValueError("Coordinates cannot be empty.")

        # if input is an object array, it might need extra help
        if coord_arr.dtype is np.dtype(object):
            first = coord_arr.ravel()[0]
            if isinstance(first, Point):
                # this will cause Points to be unpacked into numerical array
                coord_arr = np.array(coord_arr.tolist())
            else:
                raise TypeError(f"Object array with item type {type(first)} not supported as input.")
        else:
            first = coordinates
            for _ in range(coord_arr.ndim - 1):
                first = first[0]

        if isinstance(first, Point):
            source_system = first.system
        # If input was already PointArray, use its system
        elif isinstance(coordinates, PointArray):
            source_system = coordinates.system
        else:
            source_system = None

        return coord_arr, source_system

    def _check_point_operand(self, a):
        if not isinstance(a, PointArray):
            raise TypeError(f"Operand must be a PointArray or Point (received {type(a)})")
        if a.system is not self.system:
            raise ValueError(f"Operand system '{a.system}' does not match this PointArray's system '{self.system}'")

    def _check_vector_operand(self, a):
        if not isinstance(a, (Vector, VectorArray)):
            raise TypeError(f"Operand must be a VectorArray or Vector (received {type(a)})")
        if a.system is not self.system:
            raise ValueError(f"Operand system '{a.system}' does not match this PointArray's system '{self.system}'")


class Point(PointArray):
    """Represents a single point in space; one may request the coordinates of this point
    in any coordinate system.
    """

    def __init__(self, coordinates, system: CoordSysOrStr = None, graph: StrOrNone = None):
        coordinates = np.asarray(coordinates)
        if coordinates.ndim != 1:
            raise TypeError("Point coordinates must be 1D")
        super().__init__(coordinates[np.newaxis, :], system, graph)

    @property
    def coordinates(self):
        return self._coordinates[0]

    def __sub__(self, b: PointArray | VectorArray) -> VectorArray | PointArray:
        if isinstance(b, Point):
            self._check_point_operand(b)
            return Vector(b, self)
        elif isinstance(b, Vector):
            self._check_vector_operand(b)
            new_coords = self.coordinates - b.displacement
            return Point(new_coords, system=self.system)
        return super().__sub__(b)

    def __repr__(self):
        # Ensure coordinates are displayed correctly even if _coordinates is 2D
        coords_tuple = tuple(self.coordinates)
        return f"<{type(self).__name__} {coords_tuple} in {self.system.name}>"


class VectorArray:
    """
    Represents an N-dimensional array of vectors (displacements).

    A vector is defined by two endpoints, p1 (start) and p2 (end),
    within the same coordinate system. The displacement is p2 - p1.

    Initialization only accepts two PointArray/Point instances as endpoints.
    """

    def __init__(self, p1: PointArray | Point, p2: PointArray | Point):
        if not isinstance(p1, (PointArray, Point)) or not isinstance(p2, (PointArray, Point)):
            raise TypeError("Vector endpoints (p1, p2) must be PointArray or Point instances.")
        if p1.system is not p2.system:
            raise ValueError(f"Vector endpoints must share the same coordinate system ({p1.system} != {p2.system}).")

        # Shape check needs to compare the structural shape, ignoring the last coord dim.
        # Allow one operand to be a Point (ndim==1) to broadcast against a PointArray.
        if p1.ndim == p2.ndim and p1.shape[0] != p2.shape[0]:
            raise ValueError(
                f"Vector endpoints must have the same structural shape ({p1.shape[:-1]} != {p2.shape[:-1]})."
            )

        self._p1 = p1
        self._p2 = p2
        self._displacement = p2.coordinates - p1.coordinates
        # Cache displacement? For now, compute on demand.
        # self._displacement = self._p2.coordinates - self._p1.coordinates

    @property
    def p1(self) -> PointArray:
        """The starting points of the vectors."""
        return self._p1

    @property
    def p2(self) -> PointArray:
        """The ending points of the vectors."""
        return self._p2

    @property
    def system(self) -> CoordinateSystem:
        """The coordinate system of the vector endpoints."""
        return self._p1.system

    @property
    def displacement(self) -> np.ndarray:
        """The displacement array (p2.coordinates - p1.coordinates)."""
        # Ensure coordinates are used for subtraction
        return self._displacement

    @property
    def shape(self) -> tuple:
        """The shape of the array structure (like PointArray)."""
        return self.displacement.shape

    @property
    def ndim(self) -> int:
        """The number of dimensions of the array structure (like PointArray)."""
        return self.displacement.ndim

    @property
    def dtype(self) -> np.dtype:
        """The data type of the vector components."""
        # Displacement dtype might differ from p1/p2 if they are integers
        return self.displacement.dtype

    def __len__(self) -> int:
        """The number of vectors in the array (first dimension, like PointArray)."""
        return len(self.displacement)

    def __getitem__(self, index) -> np.ndarray:
        """Get the displacement vector(s) at the given index."""
        # Return the displacement component, not a new VectorArray slice
        return self.displacement[index]

    def __iter__(self):
        """Iterate over the displacement vectors."""
        yield from self.displacement

    def _check_vector_operand(self, other: VectorArray):
        """Validate operand for vector-vector operations."""
        if not isinstance(other, (Vector, VectorArray)):
            raise TypeError(f"Operand must be a VectorArray or Vector (received {type(other)})")
        if self.system is not other.system:
            raise ValueError(
                f"Operand system '{other.system}' does not match this VectorArray's system '{self.system}'"
            )
        # Shape check for addition requires broadcast compatibility or equality
        # Simple equality check for now, consider broadcasting later if needed.
        if self.shape[-1] != other.shape[-1]:
            raise ValueError(f"Operand shapes {self.shape} and {other.shape} are not compatible.")

    def __add__(self, other: VectorArray | PointArray) -> VectorArray | PointArray:
        """
        Add this vector array to another vector array or a point array.

        Vector + Vector = Vector (Result starts at self.p1)
        Vector + Point = Point
        """
        if isinstance(other, (Vector, VectorArray)):
            self._check_vector_operand(other)

            if isinstance(self, Vector) and isinstance(other, Vector):
                return Vector(self.p1, self.p2 + other)
            return VectorArray(self.p1, self.p2 + other)
        elif isinstance(other, PointArray):
            # Adding vector to point: Point + Vector = Point
            # Use PointArray._check_vector_operand for system check
            other._check_vector_operand(self)  # Check system match from Point's perspective
            new_coords = other.coordinates + self.displacement
            if isinstance(other, Point) and isinstance(self, Vector):
                return Point(new_coords, system=self.system)
            return PointArray(new_coords, system=self.system)
        return NotImplemented  # Let Python try other.__radd__(self)

    def mapped_to(self, system) -> "VectorArray":
        """
        Map this vector/vector array to a new coordinate system.

        This is done by mapping the start and end points of the vector(s)
        to the target system. The new vector(s) are defined by these
        newly mapped points.

        Parameters
        ----------
        system : str | CoordinateSystem
            The target coordinate system.

        Returns
        -------
        Vector | VectorArray
            A new vector or vector array in the target coordinate system.
            The type of the returned object will match the type of `self`.
        """
        p1_mapped = self.p1.mapped_to(system)
        p2_mapped = self.p2.mapped_to(system)
        if isinstance(self, Vector):
            return Vector(p1_mapped, p2_mapped)
        else:
            return VectorArray(p1_mapped, p2_mapped)

    def __repr__(self):
        # Use PointArray repr for consistency if not Vector subclass
        if type(self) is VectorArray:
            # Show structural shape and system
            return f"<{type(self).__name__} shape={self.shape} system={self.system.name}>"
        # Fallback for subclasses (like Vector) - this part will be overridden by Vector.__repr__
        return f"<{type(self).__name__} from={self.p1} to={self.p2}>"

    def __getstate__(self):
        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __eq__(self, b):
        if not isinstance(b, type(self)):
            return False
        if self.system is not b.system:
            return False
        # Check shapes of underlying points match structurally
        if self.shape != b.shape:
            return False
        # Let's compare displacements for robustness with floats
        return np.allclose(self.displacement, b.displacement)


class Vector(VectorArray):
    """
    Represents a single vector (displacement) in space.

    Defined by two Point endpoints, p1 (start) and p2 (end).
    Inherits transformation and addition logic from VectorArray.
    """

    def __init__(self, p1: Point, p2: Point):
        if not isinstance(p1, Point) or not isinstance(p2, Point):
            raise TypeError("Vector endpoints (p1, p2) must be Point instances.")
        # Base class init handles system and shape checks
        super().__init__(p1, p2)

    # Properties (p1, p2, system, displacement) are inherited

    # Ensure p1/p2 return Point type
    @property
    def p1(self) -> Point:
        """The starting point of the vector."""
        return self._p1  # Should already be a Point due to __init__ check

    @property
    def p2(self) -> Point:
        """The ending point of the vector."""
        return self._p2  # Should already be a Point due to __init__ check

    # Override shape/ndim/len to reflect single vector nature (like Point)
    @property
    def shape(self) -> tuple:
        """The shape of the coordinate array for the vector (e.g., (3,))."""
        # Return shape of the 1D coordinate array
        return self.p1.shape  # Same as p1.coordinates.shape

    @property
    def ndim(self) -> int:
        """The number of dimensions of the coordinate array (always 1 for Vector)."""
        return 1  # A single vector is 1D in terms of coordinates

    def __len__(self) -> int:
        """The number of coordinate dimensions."""
        return self.shape[0]  # Length of the 1D coordinate array

    def __repr__(self):
        disp_tuple = tuple(self.displacement)
        # Use short repr for endpoints
        p1_repr = repr(self.p1)
        p2_repr = repr(self.p2)
        return f"<{type(self).__name__} {disp_tuple} system={self.system.name} from={p1_repr} to={p2_repr}>"

    # _coorx_transform is handled by base class checking instance type
    # __add__ is handled by base class checking instance type
    # __eq__ uses base class implementation (comparing displacements)



import contextlib
import pickle

import numpy as np
import pytest
from coorx import CompositeTransform
from coorx import create_transform, Point
from coorx.coordinates import PointArray
from coorx.linear import STTransform
from coorx.systems import CoordinateSystemGraph
from pytest import raises

missing_tr = "No transform defined linking"
missing_cs = "No coordinate system named"
wrong_ndim = r"is \dD \(expected \dD\)"
wrong_system = "maps from system"
mult_impossible = "Cannot multiply transforms with different inner coordinate systems"
comp_impossible = "does not map to"


def test_coordinate_systems():
    default_graph = CoordinateSystemGraph.get_graph(None)
    # these coordinate system names are effectively global
    pt_cs1 = Point([0, 0], "2d-cs1")
    parr_cs2 = PointArray([[1, 1], [1, 0], [0, 1], [1, 1]], "2d-cs2")

    assert pt_cs1.system is default_graph.system("2d-cs1")
    assert np.all(pt_cs1.coordinates == [0, 0])
    with raises(TypeError):
        pt_cs1.mapped_to("2d-cs2")
    with raises(NameError, match=missing_cs):
        parr_cs2.mapped_to("nonexistent_cs")
    with raises(TypeError):
        Point([0, 0, 0], "2d-cs1")  # wrong ndim

    cs1_to_cs2 = STTransform(scale=[3, 2], offset=[10, 20], from_cs="2d-cs1", to_cs="2d-cs2")

    assert default_graph.transform("2d-cs1", "2d-cs2") is cs1_to_cs2
    assert default_graph.transform("2d-cs2", "2d-cs1") is cs1_to_cs2.inverse

    pt_cs2 = cs1_to_cs2.map(pt_cs1)

    with raises(TypeError, match=wrong_system):
        cs1_to_cs2.map(pt_cs2)
    with raises(TypeError, match=wrong_system):
        cs1_to_cs2.inverse.map(pt_cs1)

    assert pt_cs2.system is default_graph.system("2d-cs2")
    assert cs1_to_cs2.imap(pt_cs2).system is default_graph.system("2d-cs1")
    assert np.all(pt_cs2.coordinates == np.array([10, 20]))
    assert np.all(pt_cs1.coordinates == cs1_to_cs2.inverse.map(pt_cs2).coordinates)

    # composites and their inverses
    loop = CompositeTransform([cs1_to_cs2, cs1_to_cs2.inverse])
    assert loop.map(pt_cs1) == pt_cs1
    assert loop.inverse.map(pt_cs1) == pt_cs1

    # pickle point with CS
    assert pickle.loads(pickle.dumps(pt_cs1)) == pt_cs1

    # pickle transform with CS
    cs1_to_cs2_p = pickle.loads(pickle.dumps(cs1_to_cs2))
    assert cs1_to_cs2_p == cs1_to_cs2
    assert cs1_to_cs2_p.systems == cs1_to_cs2.systems


def test_mapped_to():
    CoordinateSystemGraph.get_graph(None)
    cs1_to_cs2 = STTransform(scale=[3, 2], offset=[10, 20], from_cs="2d-cs1", to_cs="2d-cs2")
    cs2_to_cs3 = STTransform(scale=[1, 1], offset=[0, 0], from_cs="2d-cs2", to_cs="2d-cs3")

    pt_cs1 = Point([0, 0], "2d-cs1")
    assert np.allclose(pt_cs1.mapped_to("2d-cs1"), pt_cs1)
    assert np.allclose(pt_cs1.mapped_to("2d-cs2"), cs1_to_cs2.map(pt_cs1))
    assert np.allclose(pt_cs1.mapped_to("2d-cs3"), cs2_to_cs3.map(cs1_to_cs2.map(pt_cs1)))
    assert np.allclose(pt_cs1.mapped_to("2d-cs3").mapped_to("2d-cs1"), pt_cs1)

    # inverses, too
    pt_cs3 = Point([0, 0], "2d-cs3")
    assert np.allclose(pt_cs3.mapped_to("2d-cs2"), cs2_to_cs3.inverse.map(pt_cs3))
    assert np.allclose(pt_cs3.mapped_to("2d-cs1"), cs1_to_cs2.inverse.map(cs2_to_cs3.inverse.map(pt_cs3)))


def test_coordinate_system_get_transform_to():
    """Test CoordinateSystem.get_transform_to method"""
    # Create transforms, which automatically creates coordinate systems in default graph
    cs1_to_cs2 = STTransform(scale=[2, 3], offset=[5, 10], from_cs="test_cs1", to_cs="test_cs2")
    cs2_to_cs3 = STTransform(scale=[1, 1], offset=[1, 1], from_cs="test_cs2", to_cs="test_cs3")
    
    # Get default graph and coordinate system objects
    default_graph = CoordinateSystemGraph.get_graph(None)
    cs1 = default_graph.system("test_cs1")
    cs2 = default_graph.system("test_cs2")
    cs3 = default_graph.system("test_cs3")
    
    # Test direct transform
    transform = cs1.get_transform_to(cs2)
    assert transform is cs1_to_cs2
    
    # Test inverse transform
    transform = cs2.get_transform_to(cs1)
    assert transform is cs1_to_cs2.inverse
    
    # Test pathfinding through intermediate coordinate system
    transform = cs1.get_transform_to(cs3)
    assert isinstance(transform, CompositeTransform)
    
    # Test transform from coordinate system to itself
    transform = cs1.get_transform_to(cs1)
    pt = Point([1, 2], cs1)
    mapped_pt = transform.map(pt)
    assert np.allclose(mapped_pt.coordinates, pt.coordinates)
    
    # Test with string coordinate system names
    transform = cs1.get_transform_to("test_cs2")
    assert transform is cs1_to_cs2
    
    # Test error when no path exists
    cs4 = default_graph.add_system("test_cs4", ndim=2)
    with raises(TypeError, match="No transform path from"):
        cs1.get_transform_to(cs4)


def test_coordinate_system_graph_transform():
    """Test CoordinateSystemGraph.transform method including pathfinding"""
    # Create transforms, which automatically creates coordinate systems in default graph
    cs1_to_cs2 = STTransform(scale=[2, 2, 2], offset=[1, 2, 3], from_cs="graph_cs1", to_cs="graph_cs2")
    cs2_to_cs3 = STTransform(scale=[0.5, 0.5, 0.5], offset=[0, 0, 0], from_cs="graph_cs2", to_cs="graph_cs3")
    cs3_to_cs4 = STTransform(scale=[1, 1, 1], offset=[10, 20, 30], from_cs="graph_cs3", to_cs="graph_cs4")
    
    # Get default graph and coordinate system objects
    default_graph = CoordinateSystemGraph.get_graph(None)
    cs1 = default_graph.system("graph_cs1")
    cs2 = default_graph.system("graph_cs2")
    cs3 = default_graph.system("graph_cs3")
    cs4 = default_graph.system("graph_cs4")
    
    # Test direct transform
    transform = default_graph.transform(cs1, cs2)
    assert transform is cs1_to_cs2
    
    # Test inverse transform
    transform = default_graph.transform(cs2, cs1)
    assert transform is cs1_to_cs2.inverse
    
    # Test pathfinding through one intermediate coordinate system
    transform = default_graph.transform(cs1, cs3)
    assert isinstance(transform, CompositeTransform)
    pt = Point([1, 2, 3], cs1)
    expected = cs2_to_cs3.map(cs1_to_cs2.map(pt))
    actual = transform.map(pt)
    assert np.allclose(actual.coordinates, expected.coordinates)
    assert actual.system is cs3
    
    # Test pathfinding through multiple intermediate coordinate systems
    transform = default_graph.transform(cs1, cs4)
    assert isinstance(transform, CompositeTransform)
    pt = Point([1, 2, 3], cs1)
    expected = cs3_to_cs4.map(cs2_to_cs3.map(cs1_to_cs2.map(pt)))
    actual = transform.map(pt)
    assert np.allclose(actual.coordinates, expected.coordinates)
    assert actual.system is cs4
    
    # Test identity transform (same coordinate system)
    transform = default_graph.transform(cs1, cs1)
    pt = Point([1, 2, 3], cs1)
    mapped_pt = transform.map(pt)
    assert np.allclose(mapped_pt.coordinates, pt.coordinates)
    assert mapped_pt.system is cs1
    
    # Test with string coordinate system names
    transform = default_graph.transform("graph_cs1", "graph_cs2")
    assert transform is cs1_to_cs2
    
    # Test mixed string and CoordinateSystem objects
    transform = default_graph.transform(cs1, "graph_cs2")
    assert transform is cs1_to_cs2
    transform = default_graph.transform("graph_cs1", cs2)
    assert transform is cs1_to_cs2
    
    # Test error when no path exists
    isolated_cs = default_graph.add_system("isolated_cs", ndim=3)
    with raises(TypeError, match="No transform path from"):
        default_graph.transform(cs1, isolated_cs)
    
    # Test error with nonexistent coordinate system
    with raises(NameError, match=missing_cs):
        default_graph.transform(cs1, "nonexistent_cs")


PARAMS = {
    "NullTransform": {},
    "TTransform": {"offset": (1, 1, 1)},
    "STTransform": {"scale": (2, 2, 2), "offset": (1, 2, 3)},
    "AffineTransform": {"matrix": [[0.5, 0, 0.707107], [0, 2, 0], [0.707107, 0, 0.5]], "offset": (4, 5, 6)},
    "SRT3DTransform": {"scale": (11, 11, 11), "angle": 45, "axis": (0, 1, 0), "offset": (1, 1, 1)},
    "TransposeTransform": {"axis_order": (1, 0, 2)},
    "LogTransform": {"base": (10, 10, 10)},
    "PolarTransform": {},
}


@pytest.mark.parametrize("type1", PARAMS.keys())
@pytest.mark.parametrize("type2", PARAMS.keys())
@pytest.mark.parametrize("inverse1", [False, True])
@pytest.mark.parametrize("inverse2", [False, True])
def test_transform_mapping(type1, type2, inverse1, inverse2):
    point = Point((1., 1., 1.), "cs1")
    if inverse1:
        cs2_from_cs1 = create_transform(type1, dims=(3, 3), systems=("cs2", "cs1"), **PARAMS[type1]).inverse
    else:
        cs2_from_cs1 = create_transform(type1, dims=(3, 3), systems=("cs1", "cs2"), **PARAMS[type1])
    if inverse2:
        cs3_from_cs2 = create_transform(type2, dims=(3, 3), systems=("cs3", "cs2"), **PARAMS[type2]).inverse
    else:
        cs3_from_cs2 = create_transform(type2, dims=(3, 3), systems=("cs2", "cs3"), **PARAMS[type2])

    assert str(cs2_from_cs1.map(point).system) == "cs2"

    explicitly_mapped = cs3_from_cs2.map(cs2_from_cs1.map(point))
    assert str(explicitly_mapped.system) == "cs3"

    with contextlib.suppress(NotImplementedError):  # ignore non-affine transforms here
        affine_mapped = cs3_from_cs2.as_affine().map(cs2_from_cs1.as_affine().map(point))
        assert str(affine_mapped.system) == "cs3"

    mult_mapped = (cs3_from_cs2 * cs2_from_cs1).map(point)
    assert str(mult_mapped.system) == "cs3"

    composite_mapped = CompositeTransform([cs2_from_cs1, cs3_from_cs2]).map(point)
    assert str(composite_mapped.system) == "cs3"

    with raises(TypeError, match=wrong_system):
        cs2_from_cs1.map(cs3_from_cs2.map(point))

    with raises(TypeError, match=mult_impossible):
        cs2_from_cs1 * cs3_from_cs2

    with raises(TypeError, match=comp_impossible):
        CompositeTransform([cs3_from_cs2, cs2_from_cs1])


def test_this_one_weird_situation():
    cs2_from_cs1 = create_transform("NullTransform", **{}, dims=(3, 3), systems=("cs1", "cs2"))
    cs3_from_cs2 = create_transform("SRT3DTransform", **PARAMS["SRT3DTransform"], dims=(3, 3), systems=("cs2", "cs3"))
    cs3_from_cs1 = cs3_from_cs2 * cs2_from_cs1
    assert str(cs3_from_cs1.map(Point([1, 1, 1], "cs1")).system) == "cs3"
    assert cs3_from_cs2.full_matrix.shape == (4, 4)  # just used to access it, really

    cs1_from_cs0 = create_transform("AffineTransform", **PARAMS["AffineTransform"], dims=(3, 3), systems=("cs0", "cs1"))
    cs3_from_cs0 = cs3_from_cs2 * cs2_from_cs1 * cs1_from_cs0
    assert str(cs3_from_cs0.map(Point([1, 1, 1], "cs0")).system) == "cs3"


@pytest.mark.parametrize("type1", PARAMS.keys())
@pytest.mark.parametrize("inverse1", [False, True])
@pytest.mark.parametrize("inverse2", [False, True])
def test_copy(type1, inverse1, inverse2):
    if inverse1:
        cs2_from_cs1 = create_transform(type1, **PARAMS[type1], dims=(3, 3), systems=("cs2", "cs1")).inverse
    else:
        cs2_from_cs1 = create_transform(type1, **PARAMS[type1], dims=(3, 3), systems=("cs1", "cs2"))
    copy = cs2_from_cs1.copy(from_cs="cs4")
    assert str(copy.systems[0]) == "cs4"
    assert str(copy.systems[1]) == "cs2"
    copy = cs2_from_cs1.copy(to_cs="cs5")
    assert str(copy.systems[0]) == "cs1"
    assert str(copy.systems[1]) == "cs5"
    copy = cs2_from_cs1.copy(from_cs="cs4", to_cs="cs5")
    assert str(copy.systems[0]) == "cs4"
    assert str(copy.systems[1]) == "cs5"


def test_composite_copy():
    cs2_from_cs1 = create_transform("AffineTransform", **PARAMS["AffineTransform"], dims=(3, 3), systems=("cs1", "cs2"))
    cs3_from_cs2 = create_transform("STTransform", **PARAMS["STTransform"], dims=(3, 3), systems=("cs2", "cs3"))
    cs3_from_cs1 = CompositeTransform([cs2_from_cs1, cs3_from_cs2])
    with pytest.raises(ValueError):
        cs3_from_cs1.copy(from_cs="cs4")
    with pytest.raises(ValueError):
        cs3_from_cs1.copy(to_cs="cs4")


@pytest.mark.parametrize("type1", PARAMS.keys())
def test_as_affine_systems(type1):
    xform = create_transform(type1, **PARAMS[type1], dims=(3, 3), systems=("affine1", "affine2"))
    point = Point([1, 2, 3], "affine1")
    with contextlib.suppress(NotImplementedError):  # ignore non-affine transforms here
        assert np.all(xform.as_affine().map(point) == xform.map(point))
        assert xform.as_affine().systems == xform.systems
        assert xform.inverse.as_affine().systems == xform.inverse.systems
        assert np.allclose(xform.inverse.as_affine().map(xform.map(point)), point)
        explicitly_looped = xform.inverse.as_affine().map(xform.as_affine().map(point))
        assert np.allclose(explicitly_looped, point)
        mult_loop = xform.inverse * xform
        assert np.allclose(mult_loop.as_affine().map(point), point)
        comp_loop = CompositeTransform([xform, xform.inverse])
        assert np.allclose(comp_loop.as_affine().map(point), point)


@pytest.mark.parametrize("type1", PARAMS.keys())
@pytest.mark.parametrize("inverse1", [False, True])
@pytest.mark.parametrize("inverse_composite", [False, True])
def test_composite_times_other(type1, inverse1, inverse_composite):
    pt_cs1 = Point([0., 0., 1.], "cs1")
    pt_cs3 = Point([0., 0., 0.], "cs3")
    if inverse_composite:
        cs1_from_cs2 = STTransform(scale=[3, 2, 1], offset=[10, 20, 30], from_cs="cs1", to_cs="cs2").inverse
        cs2_from_cs3 = STTransform(scale=[1, 1, 1], offset=[0, 0, -1], from_cs="cs3", to_cs="cs2")
        cs3_from_cs1 = CompositeTransform([cs2_from_cs3, cs1_from_cs2]).inverse
    else:
        cs2_from_cs1 = STTransform(scale=[3, 2, 1], offset=[10, 20, 30], from_cs="cs1", to_cs="cs2")
        cs3_from_cs2 = STTransform(scale=[1, 1, 1], offset=[0, 0, -1], from_cs="cs2", to_cs="cs3")
        cs3_from_cs1 = CompositeTransform([cs2_from_cs1, cs3_from_cs2])
    if inverse1:
        cs4_from_cs3 = create_transform(type1, **PARAMS[type1], dims=(3, 3), systems=("cs4", "cs3")).inverse
    else:
        cs4_from_cs3 = create_transform(type1, **PARAMS[type1], dims=(3, 3), systems=("cs3", "cs4"))

    assert str(cs3_from_cs1.map(pt_cs1).system) == "cs3"
    assert str(cs4_from_cs3.map(pt_cs3).system) == "cs4"

    explicitly_mapped = cs4_from_cs3.map(cs3_from_cs1.map(pt_cs1))
    assert str(explicitly_mapped.system) == "cs4"

    mult_mapped = (cs4_from_cs3 * cs3_from_cs1).map(pt_cs1)
    assert str(mult_mapped.system) == "cs4"

    composite_mapped = CompositeTransform([cs3_from_cs1, cs4_from_cs3]).map(pt_cs1)
    assert str(composite_mapped.system) == "cs4"

    with raises(TypeError, match=wrong_system):
        cs3_from_cs1.map(cs4_from_cs3.map(pt_cs1))

    if inverse_composite and (type1 != "NullTransform" or inverse1):
        comp_mult_err = mult_impossible
    else:
        comp_mult_err = comp_impossible
    with raises(TypeError, match=comp_mult_err):
        cs3_from_cs1 * cs4_from_cs3

    with raises(TypeError, match=comp_impossible):
        CompositeTransform([cs4_from_cs3, cs3_from_cs1])

    # check it works on the other side, too
    pt_cs0 = Point([0., 2., 0.], "cs0")
    cs0_to_cs1 = create_transform(type1, **PARAMS[type1], dims=(3, 3), systems=("cs0", "cs1"))

    assert str(cs0_to_cs1.map(pt_cs0).system) == "cs1"

    explicitly_mapped = cs3_from_cs1.map(cs0_to_cs1.map(pt_cs0))
    assert str(explicitly_mapped.system) == "cs3"

    mult_mapped = (cs3_from_cs1 * cs0_to_cs1).map(pt_cs0)
    assert str(mult_mapped.system) == "cs3"

    composite_mapped = CompositeTransform([cs0_to_cs1, cs3_from_cs1]).map(pt_cs0)
    assert str(composite_mapped.system) == "cs3"

    with raises(TypeError, match=wrong_system):
        cs0_to_cs1.map(cs3_from_cs1.map(pt_cs0))

    with raises(TypeError):
        cs0_to_cs1 * cs3_from_cs1

    with raises(TypeError):
        CompositeTransform([cs3_from_cs1, cs0_to_cs1])




import inspect
import threading
from typing import Callable
import weakref


class ChangeEvent:
    def __init__(self, transform, source_event=None):
        self.transform = transform
        self.source_event = source_event

    @property
    def sources(self):
        """A list of all transforms that changed leading to this event"""
        s = [self]
        if self.source_event is not None:
            s += self.source_event.sources
        return s


class CallbackRegistry:
    """A thread-safe registry for callbacks, supporting weak references."""
    def __init__(self):
        # List of (is_weakref, callback or weakref) tuples
        self._callbacks: list[tuple[bool, Callable]] = []
        self.lock = threading.Lock()

    def _real_callbacks(self):
        """Return (callback, is_weakref, maybe_weakref) tuples for all live callbacks."""
        return [(cb, is_weakref, cb_ref) for is_weakref, cb_ref in self._callbacks if (cb := (cb_ref() if is_weakref else cb_ref)) is not None]

    def add(self, cb, keep_reference, duplicates='error'):
        """Register a callback.
        If keep_reference is False, a weak reference to the callback is stored,
        allowing it to be garbage collected if there are no other references.

        Parameters
        ----------
        cb : Callable
            The callback to register.
        keep_reference : bool
            Whether to keep a strong reference to the callback.
        duplicates : {'error', 'ignore', 'add'}
            How to handle duplicate registrations of the same callback.
            'error' raises a ValueError, 'ignore' does nothing, 'add' registers
            the callback multiple times.
        """
        if keep_reference:
            cb_ref = (False, cb)
        else:
            # Create a cleanup call back so dead weakrefs are removed from the registry
            weak_self = weakref.ref(self)
            def cleanup(dead_ref):
                registry = weak_self()
                if registry is not None:
                    registry.remove(dead_ref)

            # create weak reference to the callback
            if inspect.ismethod(cb):
                cb_ref = (True, weakref.WeakMethod(cb, cleanup))
            else:
                cb_ref = (True, weakref.ref(cb, cleanup))

        with self.lock:
            # Prevent duplicate registrations
            if duplicates == 'allow':
                # Add the new callback regardless of previous registrations
                self._callbacks.append(cb_ref)
            else:
                # check if we have registered this callback already
                already_registered = False
                is_weakref = False
                for other_cb, is_weakref, _ in self._real_callbacks():
                    if other_cb == cb:
                        already_registered = True
                        break
                # if already registered, handle according to duplicates policy
                if already_registered:
                    if duplicates == 'error':
                        raise ValueError("Callback already registered")
                    elif duplicates == 'ignore':
                        if keep_reference and is_weakref:
                            # Upgrade weak reference to strong reference
                            self.remove(cb)
                            self._callbacks.append(cb_ref)
                        else:
                            # just ignore
                            return
                else:
                    self._callbacks.append(cb_ref)


    def remove(self, cb):
        with self.lock:
            new_callbacks = []
            for other_cb, is_ref, maybe_cb in self._real_callbacks():
                if other_cb == cb:
                    continue
                new_callbacks.append((is_ref, maybe_cb))
            self._callbacks = new_callbacks

    def __call__(self, *args, **kwargs):
        """Invoke all registered callbacks with the given arguments."""
        with self.lock:
            # Make a snapshot of callbacks to invoke
            callbacks = self._real_callbacks()
        for cb, _, _ in callbacks:
            if cb is not None:
                cb(*args, **kwargs)



import json
import math
import pickle
import unittest

import numpy as np

import coorx
from coorx import LogTransform


try:
    import itk

    HAVE_ITK = True
except ImportError:
    HAVE_ITK = False


try:
    import vispy

    HAVE_VISPY = True
except ImportError:
    HAVE_VISPY = False


try:
    import pyqtgraph as pg

    HAVE_PG = True
except ImportError:
    HAVE_PG = False


NT = coorx.NullTransform
TT = coorx.TTransform
XT = coorx.TransposeTransform
ST = coorx.STTransform
AT = coorx.AffineTransform
RT = coorx.AffineTransform
PT = coorx.PolarTransform
LT = coorx.LogTransform
CT = coorx.CompositeTransform
T3 = coorx.SRT3DTransform
ZT = coorx.AxisSelectionEmbeddedTransform
HT = coorx.HomogeneousEmbeddedTransform
DT = coorx.LensDistortionTransform
QT = coorx.PerspectiveTransform
BT = coorx.BilinearTransform
T2 = coorx.Homography2DTransform


def assert_composite_types(composite, types):
    assert list(map(type, composite.transforms)) == types


def assert_composite_objects(composite1, composite2):
    assert composite1.transforms == composite2.transforms


def test_pickling():
    points = np.random.normal(size=(10, 3))
    transforms = [
        NT(dims=(3, 3)),
        TT(dims=(3, 3), offset=(1, 2, 3)),
        ST(dims=(3, 3), scale=(2, 3, 4), offset=(1, 2, 3)),
        AT(dims=(3, 3)),
        PT(dims=(3, 3)),
        XT(axis_order=(2, 1, 0)),
        LT(dims=(3, 3), base=(10, None, 2)),
        CT([ST(dims=(3, 3), scale=(2, 2, 2)), TT(dims=(3, 3), offset=(1, 1, 1))]),
        T3(dims=(3, 3), scale=(2, 2, 2), offset=(1, 2, 3), angle=45, axis=(0, 0, 1)),
        ZT(axes=(0, 2), transform=ST(dims=(2, 2), scale=(3, 3), offset=(1, 1)), dims=(3, 3)),
        HT(transform=AT(dims=(4, 4)), dims=(3, 3)),
        QT(
            dims=(3, 3),
            affine={
                'type': 'AffineTransform',
                'matrix': [[1, 0, 0, 0], [0, 1, 0, 0], [0.001, 0.002, 1, 0], [0, 0, 0, 1]],
                'offset': [0, 0, 0, 0],
            },
        ),
        # DT(dims=(2, 2), coeff=(0.1, -0.05, 0.01, 0.0, 0.0)),
        # BT(dims=(2, 2)),
        # T2(dims=(2, 2), src_quad=[(0, 0), (1, 0), (1, 1), (0, 1)], dst_quad=[(0, 0), (2, 0), (1, 1), (0, 2)]),
    ]
    for tr in transforms:
        tr2 = pickle.loads(pickle.dumps(tr))
        assert tr == tr2
        mapped1 = tr.map(points)
        mapped2 = tr2.map(points)
        assert_equal_including_nans(mapped1, mapped2, f"Mapping differs for pickled {type(tr)}")

        alt_tr2 = coorx.create_transform(**json.loads(json.dumps(tr.save_state())))
        assert tr == alt_tr2
        mapped2b = alt_tr2.map(points)
        assert_equal_including_nans(
            mapped1, mapped2b, f"Mapping differs for create_transform'd {type(tr)}"
        )

        try:
            tr.inverse.map(tr.map(points))
        except (np.linalg.LinAlgError, NotImplementedError):
            continue  # non-invertible

        inv_tr2 = pickle.loads(pickle.dumps(tr.inverse))
        assert tr.inverse == inv_tr2
        inv_mapped1 = tr.inverse.map(mapped1)
        inv_mapped2 = inv_tr2.map(mapped2)
        assert_equal_including_nans(
            inv_mapped1, inv_mapped2, f"Inverse mapping differs for pickled {type(tr)}"
        )

        alt_inv_tr2 = coorx.create_transform(**eval(str(tr.inverse.save_state())))
        assert tr.inverse == alt_inv_tr2
        alt_inv_mapped2 = alt_inv_tr2.map(mapped2)
        assert_equal_including_nans(
            inv_mapped1,
            alt_inv_mapped2,
            f"Inverse mapping differs for create_transform'd {type(tr)}",
        )


def assert_equal_including_nans(a, b, msg):
    assert np.all(
        (np.isnan(a) == np.isnan(b)) & (np.isclose(a, b) | (np.isnan(a) & np.isnan(b)))
    ), msg
    mask = ~np.isnan(a)
    assert np.allclose(a[mask], b[mask]), msg


class TransformMultiplication(unittest.TestCase):
    def test_multiplication(self):
        n = NT(dims=(3, 3))
        t = TT(dims=(3, 3))
        s = ST(dims=(3, 3))
        a = AT(dims=(3, 3))
        p = PT(dims=(3, 3))
        x = XT(axis_order=(2, 1, 0))
        log_trans = LT(dims=(3, 3))
        c1 = CT([s, a, p, x])
        assert c1
        c2 = CT([s, a, s])

        assert isinstance(n * n, NT)
        assert isinstance(n * t, TT)
        assert isinstance(n * s, ST)
        assert isinstance(n * p, PT)
        assert isinstance(t * t, TT)
        assert isinstance(x * x, XT)
        assert isinstance(t * s, ST)
        assert isinstance(t * a, AT)
        assert isinstance(s * t, ST)
        assert isinstance(s * s, ST)
        assert isinstance(s * a, AT)
        assert isinstance(s * p, CT)
        assert isinstance(a * t, AT)
        assert isinstance(a * s, AT)
        assert isinstance(a * a, AT)
        assert isinstance(a * p, CT)
        assert isinstance(p * a, CT)
        assert isinstance(p * s, CT)
        assert isinstance(s * x, CT)
        assert isinstance(x * a, CT)
        assert isinstance(t * x, CT)
        assert isinstance(x.inverse * t, CT)
        assert_composite_types(p * a, [AT, PT])
        assert_composite_types(p * s, [ST, PT])
        assert_composite_types(s * p, [PT, ST])
        assert_composite_types(s * p * a, [AT, PT, ST])
        assert_composite_types(s * a * p, [PT, AT])
        assert_composite_types(p * s * a, [AT, ST, PT])
        assert_composite_types(s * p * s, [ST, PT, ST])
        assert_composite_types(s * a * p * x * s * a, [AT, ST, XT, PT, AT])
        assert_composite_types(c2 * a, [AT, ST, AT, ST])
        assert_composite_types(p * log_trans * s, [ST, LT, PT])


class VectorMapping(unittest.TestCase):
    @unittest.skipIf(not HAVE_PG, "pyqtgraph could not be imported")
    def test_vector_map(self):
        vector = pg.Vector(1, 2, 3)
        t = coorx.STTransform(scale=(2, 3, 4), offset=(5, 6, 7))
        mapped = t.map(vector)
        assert isinstance(mapped, pg.Vector)
        self.assertEqual(mapped.x(), 1 * 2 + 5)
        self.assertEqual(mapped.y(), 2 * 3 + 6)
        self.assertEqual(mapped.z(), 3 * 4 + 7)


class CompositeTransform(unittest.TestCase):
    def test_transform_composite(self):
        # Make dummy classes for easier distinguishing the transforms

        class DummyTrans(coorx.Transform):
            def _validate_dims(self, dims, **kwargs):
                return dims

        class TransA(DummyTrans):
            pass

        class TransB(DummyTrans):
            pass

        class TransC(DummyTrans):
            pass

        # Create test transforms
        a, b, c = TransA(), TransB(), TransC()

        # Test Composite creation
        assert coorx.CompositeTransform().transforms == []
        assert coorx.CompositeTransform([a]).transforms == [a]
        assert coorx.CompositeTransform([a, b]).transforms == [a, b]
        assert coorx.CompositeTransform([a, b, c, a]).transforms == [a, b, c, a]

        # Test composition by multiplication
        assert_composite_objects(a * b, coorx.CompositeTransform([b, a]))
        assert_composite_objects(a * b * c, coorx.CompositeTransform([c, b, a]))
        assert_composite_objects(a * b * c * a, coorx.CompositeTransform([a, c, b, a]))

        # Test adding/prepending to transform
        composite = coorx.CompositeTransform()
        composite.append(a)
        assert composite.transforms == [a]
        composite.append(b)
        assert composite.transforms == [a, b]
        composite.append(c)
        assert composite.transforms == [a, b, c]
        composite.prepend(b)
        assert composite.transforms == [b, a, b, c]
        composite.prepend(c)
        assert composite.transforms == [c, b, a, b, c]

        # Test simplifying
        t1 = coorx.STTransform(scale=(2, 3))
        t2 = coorx.STTransform(offset=(3, 4))
        t3 = coorx.STTransform(offset=(3, 4))
        # Create multiplied versions
        t123 = t1 * t2 * t3
        t321 = t3 * t2 * t1
        c123 = coorx.CompositeTransform([t3, t2, t1])
        c321 = coorx.CompositeTransform([t1, t2, t3])
        c123s = c123.simplified
        c321s = c321.simplified

        assert isinstance(t123, coorx.STTransform)
        assert isinstance(t321, coorx.STTransform)
        assert isinstance(c123s, coorx.CompositeTransform)
        assert isinstance(c321s, coorx.CompositeTransform)

        # Test Mapping
        t1 = coorx.STTransform(scale=(2, 3))
        t2 = coorx.STTransform(offset=(3, 4))
        composite12 = coorx.CompositeTransform([t1, t2])
        composite21 = coorx.CompositeTransform([t2, t1])

        assert composite12.transforms == [t1, t2]
        assert composite21.transforms == [t2, t1]

        m12 = (t2 * t1).map((1, 1)).tolist()
        m21 = (t1 * t2).map((1, 1)).tolist()
        m12_ = composite12.map((1, 1)).tolist()
        m21_ = composite21.map((1, 1)).tolist()

        assert m12 != m21
        assert m12 == m12_
        assert m21 == m21_

        # test pickle
        s = pickle.dumps(composite12)
        assert pickle.loads(s) == composite12

    def test_srt_composites(self):
        s = coorx.STTransform(scale=(2, 3, 1))
        t = coorx.TTransform(offset=(3, 4, 5))
        srt = coorx.SRT3DTransform(scale=(2, 3, 1), offset=(3, 4, 5), angle=90, axis=(0, 0, 1))

        assert isinstance(srt, coorx.SRT3DTransform)
        assert isinstance(s * t, coorx.STTransform)
        assert isinstance(t * s, coorx.STTransform)
        assert isinstance(s * srt, coorx.CompositeTransform)
        assert isinstance(srt * s, coorx.CompositeTransform)
        assert isinstance(t * srt, coorx.CompositeTransform)
        assert isinstance(srt * t, coorx.CompositeTransform)

    def test_inverse_composite(self):
        # Test inverse of composite
        t1 = coorx.STTransform(scale=(2, 3))
        t2 = coorx.STTransform(offset=(3, 4))
        composite = coorx.CompositeTransform([t1, t2])
        composite_inv = composite.inverse

        assert composite_inv.map(composite.map((1, 1))).tolist() == [1, 1]

    def test_map_order(self):
        t1 = coorx.SRT3DTransform(scale=(2, 3, 1), offset=(3, 4, 5), angle=90, axis=(0, 0, 1))
        t2 = coorx.TTransform(offset=(3, 4, 5))
        t3 = coorx.TTransform(offset=(13, 4.1, 7)).inverse
        composite = t1 * t2 * t3
        explicit = coorx.CompositeTransform([t3, t2, t1])
        assert composite.map((1, 1, 1)).tolist() == t1.map(t2.map(t3.map((1, 1, 1)))).tolist()
        assert composite.map((1, 1, 1)).tolist() == explicit.map((1, 1, 1)).tolist()

        composite = t3 * t2 * t1
        explicit = coorx.CompositeTransform([t1, t2, t3])
        assert composite.map((1, 1, 1)).tolist() == t3.map(t2.map(t1.map((1, 1, 1)))).tolist()
        assert composite.map((1, 1, 1)).tolist() == explicit.map((1, 1, 1)).tolist()

    def test_as_affine(self):
        t1 = coorx.STTransform(scale=(2, 3, 1))
        t2 = coorx.TTransform(offset=(3, 4, 0))
        t3 = coorx.AffineTransform(dims=(3, 3))
        t3.rotate(90, (0, 0, 1))
        t3 = t3.inverse
        t4 = coorx.NullTransform(dims=3)
        composite = coorx.CompositeTransform([t1, t2, t3, t4])
        affine = composite.as_affine()
        assert isinstance(affine, coorx.AffineTransform)

    def test_full_matrix(self):
        t1 = coorx.STTransform(scale=(2, 3, 5))
        t2 = coorx.STTransform(offset=(3, 4, 2))
        composite = coorx.CompositeTransform([t1, t2.inverse])
        assert np.allclose(composite.full_matrix, np.dot(t2.inverse.full_matrix, t1.full_matrix))

    def test_save_state(self):
        t1 = coorx.STTransform(scale=(2, 3, 5))
        t2 = coorx.STTransform(offset=(3, 4, 0))
        matrix = np.eye(3)
        matrix[:2, :2] = [[0, -0.1], [0.1, 0]]
        t3 = coorx.AffineTransform(matrix, offset=(1, 2, 3))
        data = np.random.normal(size=(10, 3))
        composite = coorx.CompositeTransform([t1, t2, t3])
        state = eval(str(composite.save_state()))
        assert state['type'] == 'CompositeTransform'

        composite2 = coorx.create_transform(**state)
        assert composite == composite2
        assert np.allclose(composite.map(data), composite2.map(data))

    @unittest.skipIf(not HAVE_VISPY, "vispy could not be imported")
    def test_as_vispy(self):
        from vispy.visuals.transforms import ChainTransform

        t1 = coorx.STTransform(scale=(2, 3, 5))
        t2 = coorx.STTransform(offset=(3, 4, 0))
        composite = coorx.CompositeTransform([t1, t2])
        as_vispy = composite.as_vispy()
        assert isinstance(as_vispy, ChainTransform)
        assert np.allclose(as_vispy.map((1, 1, 1))[:3], composite.map((1, 1, 1)))


class TTransform(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        self.transforms = [
            TT(dims=(3, 3)),
            TT([1]),
            TT([0, 0]),
            TT(np.array([1, 1e16])),
            TT((-100e-6, 12e8, 0)),
            TT(np.random.normal(size=10)),
        ]
        self.points = [
            np.random.normal(size=(10, 3)),
            10 ** np.random.normal(size=(10, 3)),
            -(10 ** np.random.normal(size=(10, 3))),
            [2, 3, 4],
            [10],
            (5, 6),
            [[0, 0, 0], [1, 0, 0], [0, 1, 0]],
            [(0, 0, 0), (1, 0, 0), (0, 1, 0)],
        ]

    def test_t_transform(self):
        # Check that TTransform maps exactly like AffineTransform
        pts = np.random.normal(size=(10, 3))
        translate = (1e6, 0.2, 0)
        mapped = pts + translate
        imapped = pts - translate
        tt = coorx.TTransform(offset=translate)

        assert np.allclose(tt.map(pts), mapped)
        assert np.allclose(tt.inverse.map(pts), imapped)

        # test save/restore
        tt2 = coorx.TTransform(dims=(3, 3))
        tt2.__setstate__(tt.__getstate__())
        assert np.allclose(tt.map(pts), tt2.map(pts))

        tt3 = pickle.loads(pickle.dumps(tt))
        assert np.allclose(tt.map(pts), tt3.map(pts))

    @unittest.skipIf(not HAVE_ITK, "itk could not be imported")
    def test_itk_compat(self):
        itk_tr = itk.TranslationTransform[itk.D, 3].New()
        ttr = TT(dims=(3, 3))

        pts = 10 ** np.random.normal(size=(20, 3), scale=16)
        offsets = 10 ** np.random.normal(size=(20, 3), scale=16)

        for offset in offsets:
            ttr_pts = ttr.map(pts)
            for i in range(len(pts)):
                itk_tr_pt = np.array(itk_tr.TransformPoint(itk.Point[itk.D, 3](pts[i])))
                assert np.allclose(itk_tr_pt, ttr_pts[i])
            ttr.translate(offset)
            itk_tr.Translate(itk.Point[itk.D, 3](offset))
            assert np.allclose(ttr.offset, np.array(itk_tr.GetOffset()))

    def test_inverse(self):
        for tr in self.transforms:
            for pts in self.points:
                arr = np.array(pts)
                if arr.shape[-1] == tr.dims[0]:
                    pts2 = tr.map(pts)
                    assert pts2.shape[-1] == tr.dims[1]
                    pts3 = tr.imap(pts2)
                    assert np.allclose(arr, pts3)
                else:
                    with self.assertRaises(TypeError):
                        tr.map(pts)


class STTransform(unittest.TestCase):
    def test_st_transform(self):
        # Check that STTransform maps exactly like AffineTransform
        pts = np.random.normal(size=(10, 3))

        scale = (1, 7.5, -4e-8)
        translate = (1e6, 0.2, 0)
        st = coorx.STTransform(scale=scale, offset=translate)
        at = coorx.AffineTransform(dims=(3, 3))
        at.zoom(scale)
        at.translate(translate)

        assert np.allclose(st.map(pts), at.map(pts))
        assert np.allclose(st.inverse.map(pts), at.inverse.map(pts))

    def test_st_mapping(self):
        p1 = [[5.0, 7.0], [23.0, 8.0]]
        p2 = [[-1.3, -1.4], [1.1, 1.2]]

        t = coorx.STTransform(dims=(2, 2))
        t.set_mapping(p1, p2)

        assert np.allclose(t.map(p1)[:, : len(p2)], p2)


def check_matrix(t, m):
    dim = m.shape[0] - 1
    assert np.allclose(t.full_matrix, m)
    assert np.allclose(t.matrix, m[:dim, :dim])
    assert np.allclose(t.offset, m[:dim, dim])


class AffineTransform(unittest.TestCase):
    def test_modifiers(self):
        t = coorx.AffineTransform(dims=(3, 3))
        m = np.eye(4)
        check_matrix(t, m)

        t.translate(1)
        m[:3, 3] += 1
        check_matrix(t, m)

        t.zoom(2)
        m[:3] *= 2
        check_matrix(t, m)

        t.translate(3)
        m[:3, 3] += 3
        check_matrix(t, m)

        t.rotate(90, (0, 0, 1))
        m[:2] = [[0, 2, 0, 5], [-2, 0, 0, -5]]
        check_matrix(t, m)

        rm = coorx.AffineTransform(dims=(3, 3))
        rm.rotate(90, (0, 0, 1))
        t2 = (
            rm
            * coorx.TTransform([3, 3, 3])
            * coorx.STTransform(scale=[2, 2, 2])
            * coorx.TTransform([1, 1, 1])
        )
        assert t2 == t

    def test_4d(self):
        t = coorx.AffineTransform(dims=(4, 4))
        m = np.eye(5)
        check_matrix(t, m)

        t.translate([1, 2, 3, 4])
        m[:4, 4] += [1, 2, 3, 4]
        check_matrix(t, m)

        t.zoom([2, 3, 4, 5])
        for i in range(4):
            m[i] *= [2, 3, 4, 5][i]
        check_matrix(t, m)

        t.translate([5, 6, 7, 8])
        m[:4, 4] += [5, 6, 7, 8]
        check_matrix(t, m)

        inv = t.inverse
        a = np.random.normal(size=(10, 4))
        assert np.allclose(inv.map(t.map(a)), a)

    def test_changing_dims(self):
        matrix = [[1, 0, 1], [0, 1, 1]]
        with self.assertRaises(ValueError):
            t = AT(matrix=matrix, dims=3)
        with self.assertRaises(ValueError):
            t = AT(matrix=matrix, dims=(2, 2))
        with self.assertRaises(ValueError):
            t = AT(matrix=matrix, offset=[1, 2, 3])

        t = AT(matrix=matrix, offset=[1, 2])
        assert t.dims == (3, 2)

        with self.assertRaises(ValueError):
            t.matrix = np.eye(3)

        pts = np.random.normal(size=(10, 3))
        mapped = t.map(pts)
        assert mapped.shape == (10, 2)

        with self.assertRaises(np.linalg.LinAlgError):
            t.imap(mapped)

    def test_inverse(self):
        tr = AT(dims=(3, 3), matrix=np.eye(3) * 0.1, offset=[1, 2, 3])
        pts = np.random.normal(size=(10, 3))
        arr = np.array(pts)
        if arr.shape[-1] == tr.dims[0]:
            pts2 = tr.map(pts)
            assert pts2.shape[-1] == tr.dims[1]
            pts3 = tr.imap(pts2)
            assert np.allclose(arr, pts3)
        else:
            with self.assertRaises(TypeError):
                tr.map(pts)

    def test_uninvertible(self):
        tr = AT(dims=(2, 2), matrix=[[1, 1], [2, 2]], offset=[0, 0])
        pts = np.random.normal(size=(10, 2))
        with self.assertRaises(np.linalg.LinAlgError):
            tr.imap(pts)
        with self.assertRaises(np.linalg.LinAlgError):
            tr.inverse.map(pts)

    def x_test_affine_mapping(self):
        t = coorx.AffineTransform()
        p1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])

        # test pure translation
        p2 = p1 + 5.5
        t.set_mapping(p1, p2)
        assert np.allclose(t.map(p1)[:, : p2.shape[1]], p2)
        t2 = coorx.AffineTransform()
        t2.translate(5.5)
        assert np.allclose(t.full_matrix, t2.full_matrix)

        # test pure scaling
        p2 = p1 * 5.5
        t.set_mapping(p1, p2)
        assert np.allclose(t.map(p1)[:, : p2.shape[1]], p2)
        t2 = coorx.AffineTransform()
        t2.zoom(5.5)
        assert np.allclose(t.full_matrix, t2.full_matrix)

        # test scale + translate
        p2 = (p1 * 5.5) + 3.5
        t.set_mapping(p1, p2)
        assert np.allclose(t.map(p1)[:, : p2.shape[1]], p2)
        t2 = coorx.AffineTransform()
        t2.zoom(3.5)
        t2.translate(5.5)
        assert np.allclose(t.full_matrix, t2.full_matrix)

        # test scale + translate + rotate
        p2 = np.array([[10, 5, 3], [10, 15, 3], [30, 5, 3], [10, 5, 3.5]])
        t.set_mapping(p1, p2)
        assert np.allclose(t.map(p1)[:, : p2.shape[1]], p2)
        t2 = coorx.AffineTransform()
        t2.zoom(3.5)
        t2.rotate(90)
        t2.translate(5.5)
        assert np.allclose(t.full_matrix, t2.full_matrix)

    def test_to_and_from_qmatrix4x4(self):
        """Test conversion to and from QMatrix4x4 objects."""
        # Skip test if no Qt is available
        try:
            from coorx.qt import import_qt_gui

            import_qt_gui()
        except ImportError:
            self.skipTest("Qt GUI module not available")

        # Create a test transform with rotation, scaling, and translation
        tr = coorx.AffineTransform(dims=(3, 3))
        tr.zoom((2, 3, 0.5))
        tr.rotate(45, (0, 0, 1))
        tr.translate((10, 20, 30))

        # Convert to QMatrix4x4 and back
        qmatrix = tr.to_qmatrix4x4()
        tr2 = coorx.AffineTransform.from_qmatrix4x4(qmatrix)

        # Check that full matrices are identical
        assert np.allclose(tr.full_matrix, tr2.full_matrix)

        # Check that transforms produce identical results
        pts = np.random.normal(size=(10, 3))
        assert np.allclose(tr.map(pts), tr2.map(pts))


class TransposeTransformTest(unittest.TestCase):
    def test_inverse_and_map(self):
        tt = XT(axis_order=(2, 1, 0))
        pts = np.random.normal(size=(10, 3))
        assert np.allclose(tt.inverse.map(tt.map(pts)), pts)
        assert np.allclose(tt.map(pts), pts[..., ::-1])


class LogTransformTest(unittest.TestCase):
    def test_log(self):
        lt = LogTransform((12, None))  # None is the identity value
        data = [(12, -6), (144, 13.2), (float("inf"), 21), (-7, 44), (0, 0)]
        output = lt.map(data)
        self.assertAlmostEqual(output[0][0], 1)
        self.assertAlmostEqual(output[1][0], 2)
        self.assertAlmostEqual(output[1][1], 13.2)  # None = identity, so unchanged
        self.assertTrue(math.isnan(output[2][0]))
        self.assertTrue(math.isnan(output[3][0]))
        self.assertTrue(math.isnan(output[4][0]))
        self.assertAlmostEqual(output[4][1], 0)  # None = identity, so unchanged


class SRT3DTransformTest(unittest.TestCase):
    def test_srt3d(self):
        pts = np.random.normal(size=(10, 3))

        tr = coorx.SRT3DTransform()
        aff = coorx.AffineTransform(dims=(3, 3))
        assert np.allclose(pts, tr.map(pts))

        scale = [10, 1, 0.1]
        tr.scale = scale
        aff.zoom(scale)
        assert np.allclose(aff.map(pts), tr.map(pts))

        angle = 30
        axis = np.array([1, 0.5, 0.3])
        tr.rotation = angle, axis
        aff.rotate(angle, axis)
        assert np.allclose(aff.map(pts), tr.map(pts))
        assert tr.rotation[0] == angle
        assert np.allclose(tr.rotation[1], axis)

        offset = [1e-6, -10, 1e6]
        tr.offset = offset
        aff.translate(offset)
        assert np.allclose(aff.map(pts), tr.map(pts))

    @unittest.expectedFailure
    def test_from_affine(self):
        # TODO this was previously part of test_srt3d, but it breaks trying to initialize tr2
        tr2 = coorx.SRT3DTransform(init=aff)
        assert np.allclose(tr.offset, tr2.offset)
        assert np.allclose(tr.scale, tr2.scale)
        assert tr.rotation[0] == tr2.rotation[0]
        assert np.allclose(tr.rotation[1], tr2.rotation[1])
        assert np.allclose(tr2.map(pts), tr.map(pts))

    def test_save(self):
        tr = coorx.SRT3DTransform(scale=(1, 2, 3), offset=(10, 5, 3), angle=120, axis=(1, 1, 2))
        s = eval(str(tr.save_state()))
        assert s["type"] == "SRT3DTransform"
        assert tuple(s["dims"]) == (3, 3)

    @unittest.skipIf(not HAVE_VISPY, "vispy could not be imported")
    def test_as_vispy(self):
        tr = coorx.SRT3DTransform(scale=(1, 2, 3), offset=(10, 5, 3), angle=120, axis=(1, 1, 2))
        vt = tr.as_vispy()
        assert np.allclose(vt.map((1, 1, 1))[:3], tr.map((1, 1, 1)))
        assert np.allclose(vt.map((1, 3, 5))[:3], tr.map((1, 3, 5)))

    @unittest.skipIf(not HAVE_PG, "pyqtgraph could not be imported")
    def test_to_and_from_pyqtgraph(self):
        axis = np.array((1, 1, 2))
        axis = axis / np.linalg.norm(axis)
        tr = coorx.SRT3DTransform(scale=(1, 2, 3), offset=(10, 5, 3), angle=120, axis=axis)
        tr2 = coorx.SRT3DTransform.from_pyqtgraph(tr.as_pyqtgraph())
        assert np.allclose(tr.full_matrix, tr2.full_matrix)
        pt = np.random.normal(size=(10, 3))
        assert np.allclose(tr.map(pt), tr2.map(pt))

    def test_composite(self):
        tr1 = coorx.SRT3DTransform(offset=(1, 2, 3))
        tr2 = coorx.SRT3DTransform(scale=(10, 10, 1))

        # test multiplication
        tr3 = tr1 * tr2  # scale, then offset
        assert isinstance(tr3, coorx.AffineTransform)
        assert np.allclose(tr3.map([0, 0, 0]), [1, 2, 3])
        assert np.allclose(tr3.map([1, 1, 1]), [11, 12, 4])

        # test composite
        tr4 = coorx.CompositeTransform([tr2, tr1])
        assert len(tr4.simplified.transforms) == 1
        assert isinstance(tr4.simplified.transforms[0], coorx.AffineTransform)
        assert np.all(tr3.matrix == tr4.simplified.transforms[0].matrix)
        assert np.all(tr3.offset == tr4.simplified.transforms[0].offset)
        assert np.all(tr3.full_matrix == tr4.simplified.transforms[0].full_matrix)
        assert np.allclose(tr4.map([2, -26.7, 0]), tr3.map([2, -26.7, 0]))

    def test_composite_setitem(self):
        tr1 = coorx.SRT3DTransform(offset=(1, 2, 3))
        tr2 = coorx.SRT3DTransform(scale=(10, 10, 1))
        tr3 = coorx.SRT3DTransform(scale=(1, 11, 111))

        tr4 = coorx.CompositeTransform([tr2, tr1])
        tr4[0] = tr3
        assert np.all(tr4.transforms[0].full_matrix == tr3.full_matrix)


class TransformInverse(unittest.TestCase):
    def test_inverse(self):
        m = np.random.normal(size=(3, 3))
        transforms = [
            NT(dims=3),
            ST(scale=(1e-4, 2e5, 1), offset=(10, -6e9, 0)),
            AT(m),
            RT(m),
        ]

        np.random.seed(0)
        N = 20
        x = np.random.normal(size=(N, 3))
        pw = np.random.normal(size=(N, 3), scale=3)
        pos = x * 10**pw

        for trn in transforms:
            assert np.allclose(pos, trn.inverse.map(trn.map(pos))[:, :3])

        # log transform only works on positive values
        # abs_pos = np.abs(pos)
        # tr = LT(base=(2, 4.5, 0))
        # assert np.allclose(abs_pos, tr.inverse.map(tr.map(abs_pos))[:,:3])


class BilinearTest(unittest.TestCase):
    def test_bilinear(self):
        # identity
        tr = self.check_mapping([[0, 0], [1, 0], [0, 1], [1, 1]], [[0, 0], [1, 0], [0, 1], [1, 1]])
        assert np.allclose(tr.matrix, np.eye(4)[1:3])

        tr = self.check_mapping([[0, 0], [1, 0], [0, 1], [1, 1]], [[0, 0], [1, 1], [0, 1], [1, 0]])

        tr = self.check_mapping([[0, 0], [1, 0], [0, 1], [1, 1]], [[1, 1], [3, 0], [0, 4], [7, 7]])

    def check_mapping(self, a, b):
        a = np.asarray(a)
        b = np.asarray(b)

        tr = coorx.linear.BilinearTransform()
        tr.set_mapping(a, b)
        c = tr.map(a)
        assert c.shape == b.shape
        assert np.allclose(c, b)

        assert np.allclose(tr.map(a.mean(axis=0)), b.mean(axis=0))

        tr_inv = tr.inverse
        d = tr_inv.map(b)
        assert d.shape == a.shape
        assert np.allclose(d, a)

        # Note: there are cases where a bilinear transform is not invertible
        # assert np.allclose(
        #     tr_inv.map(b.mean(axis=0)),
        #     a.mean(axis=0)
        # )

        assert np.all(tr_inv.map(b) == tr.imap(b))

        return tr
